import random
import numpy as np
import torch
from torch.utils.data import DataLoader
from lightwood.mixer.helpers.ranger import Ranger
from lightwood.encoder.categorical.simple_label import SimpleLabelEncoder
from lightwood.encoder.categorical.onehot import OneHotEncoder
from lightwood.encoder.categorical.gym import Gym
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log
from lightwood.mixer.helpers.default_net import DefaultNet
import pandas as pd

from typing import Iterable, Tuple, List


class CategoricalAutoEncoder(BaseEncoder):
    """
    Trains an autoencoder (AE) to represent categorical information with over 100 categories. This is used to ensure that feature vectors for categorical data with many categories are not excessively large.

    The AE defaults to a vector sized 100 but can be adjusted to user preference. It is highly advised NOT to use this encoder to feature engineer your target, as reconstruction accuracy will determine your AE's ability to decode properly.

    """  # noqa

    is_trainable_encoder: bool = True

    def __init__(
            self,
            stop_after: float = 3600,
            is_target: bool = False,
            max_encoded_length: int = 100,
            desired_error: float = 0.01,
            batch_size: int = 200,
            device: str = '',
            input_encoder: str = None
    ):
        """
        :param stop_after: Stops training with provided time limit (sec)
        :param is_target: Encoder represents target class (NOT recommended)
        :param max_encoded_length: Maximum length of vector represented
        :param desired_error: Threshold for reconstruction accuracy error
        :param batch_size: Minimum batch size while training
        :param device: Name of the device that get_device_from_name will attempt to use
        :param input_encoder: one of `OneHotEncoder` or `SimpleLabelEncoder` to force usage of the underlying input encoder. Note that OHE does not scale for categorical features with high cardinality, while SLE can but is less accurate overall.
        """  # noqa
        super().__init__(is_target)
        self.is_prepared = False
        self.name = 'Categorical Autoencoder'
        self.output_size = max_encoded_length

        # Model details
        self.net = None
        self.encoder = None
        self.decoder = None
        self.input_encoder = None  # TBD at prepare()
        self.device_type = device
        self.input_encoder = input_encoder

        # Training details
        self.batch_size = batch_size
        self.desired_error = desired_error
        self.stop_after = stop_after

    def prepare(self, train_priming_data: pd.Series, dev_priming_data: pd.Series):
        """
        Creates inputs and prepares a categorical autoencoder (CatAE) for input data. Currently, does not support a dev set; inputs for train and dev are concatenated together to train an autoencoder.

        :param train_priming_data: Input training data
        :param dev_priming_data: Input dev data (Not supported currently)
        """  # noqa

        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        if self.is_target:
            log.warning('You are trying to use an autoencoder for the target value! This is very likely a bad idea.')

        error_msg = f'Provided an invalid input encoder ({self.input_encoder}), please use either `OneHotEncoder` or `SimpleLabelEncoder`.'  # noqa
        if self.input_encoder is not None:
            assert self.input_encoder in ('OneHotEncoder', 'SimpleLabelEncoder'), error_msg

        log.info('Preparing a categorical autoencoder.')

        if self.input_encoder == 'SimpleLabelEncoder' or \
                (self.input_encoder is None and train_priming_data.nunique() > 500):
            log.info('Deploying SimpleLabelEncoder for CategoricalAutoEncoder input.')
            self.input_encoder = SimpleLabelEncoder(is_target=self.is_target)
            input_len = self.input_encoder.output_size
            self.output_size = 32
            net_shape = [input_len, 128, 64, self.output_size, input_len]
        else:
            log.info('Deploying OneHotEncoder for CategoricalAutoEncoder input.')
            self.input_encoder = OneHotEncoder(is_target=self.is_target)
            net_shape = None   # defined at prepare() due to the OHE output size being determined then

        train_loader, dev_loader = self._prepare_AE_input(
            train_priming_data, dev_priming_data
        )

        best_model = self._prepare_catae(train_loader, dev_loader, net_shape=net_shape)
        self.net = best_model.to(self.net.device)

        modules = [
            module
            for module in self.net.modules()
            if type(module) != torch.nn.Sequential and type(module) != DefaultNet
        ]

        self.encoder = torch.nn.Sequential(*modules[0:-1]).eval()
        self.decoder = torch.nn.Sequential(*modules[-1:]).eval()
        log.info('Categorical autoencoder ready.')

        self.is_prepared = True

    def encode(self, column_data: Iterable[str]) -> torch.Tensor:
        """
        Encodes categorical information in column as the compressed vector from the CatAE.

        :param column_data: An iterable of category samples from a column

        :returns: An embedding for each sample in original input
        """  # noqa
        encoded_tensor = self.input_encoder.encode(column_data)

        with torch.no_grad():
            encoded_tensor = encoded_tensor.to(self.net.device)
            if len(encoded_tensor.shape) < 2:
                encoded_tensor = encoded_tensor.unsqueeze(-1)
            embeddings = self.encoder(encoded_tensor)
            return embeddings.to('cpu')

    def decode(self, encoded_data: torch.Tensor) -> List[str]:
        """
        Decodes from the embedding space, the original categories.

        ..warning If your reconstruction accuracy is not 100%, the CatAE may not return the correct category.

        :param encoded_data: A torch tensor of embeddings for category predictions

        :returns: A list of 'translated' categories for each embedding
        """  # noqa
        with torch.no_grad():
            encoded_data = encoded_data.to(self.net.device)
            encoded_tensor = self.decoder(encoded_data)
            encoded_tensor = encoded_tensor.to('cpu')
            return self.input_encoder.decode(encoded_tensor)

    def _prepare_AE_input(
            self, train_priming_data: pd.Series, dev_priming_data: pd.Series
    ) -> Tuple[DataLoader, DataLoader]:
        """
        Creates the data loaders for the CatAE model inputs. Expected inputs are generally of form `pd.Series`

        Currently does not use 'dev'; concatenates both inputs together.

        Input to the `DataLoader` must be an Iterable[str] (ideally List[str])

        """  # noqa
        if len(dev_priming_data) > 0:
            priming_data = (
                pd.concat([train_priming_data, dev_priming_data]).astype(str).tolist()
            )
        else:
            priming_data = [str(x) for x in train_priming_data]

        random.seed(len(priming_data))

        # Prepare a one-hot encoder for CatAE inputs
        self.input_encoder.prepare(priming_data)
        self.batch_size = max(min(self.batch_size, int(len(priming_data) / 50)), 1)

        train_loader = DataLoader(
            list(zip(priming_data, priming_data)),
            batch_size=self.batch_size,
            shuffle=True,
        )

        # TODO; make `Gym` compatible with a dev set
        dev_loader = None

        return train_loader, dev_loader

    def _prepare_catae(self, train_loader: DataLoader, dev_loader: DataLoader, net_shape=None):
        """
        Trains the CatAE using Lightwood's `Gym` class.

        :param train_loader: Training dataset Loader
        :param dev_loader: Validation set DataLoader
        """  # noqa
        if net_shape is None:
            input_len = self.input_encoder.output_size
            net_shape = [input_len, self.output_size, input_len]

        self.net = DefaultNet(shape=net_shape, device=self.device_type)

        if isinstance(self.input_encoder, OneHotEncoder):
            criterion = torch.nn.CrossEntropyLoss()
            desired_error = self.desired_error
        elif isinstance(self.input_encoder, SimpleLabelEncoder):
            criterion = torch.nn.MSELoss()
            desired_error = 1e-9
        else:
            raise Exception(f'[CatAutoEncoder] Input encoder of type {type(self.input_encoder)} is not supported!')

        if isinstance(self.input_encoder, OneHotEncoder):
            optimizer = Ranger(self.net.parameters())
            output_encoder = self._encoder_targets
            max_time = self.stop_after
        else:
            optimizer = Ranger(self.net.parameters(), weight_decay=1e-2)
            output_encoder = self._label_targets
            max_time = 60 * 2

        gym = Gym(
            model=self.net,
            optimizer=optimizer,
            scheduler=None,
            loss_criterion=criterion,
            device=self.net.device,
            name=self.name,
            input_encoder=self.input_encoder.encode,
            output_encoder=output_encoder,
        )

        best_model, _, _ = gym.fit(
            train_loader,
            dev_loader,
            desired_error=desired_error,
            max_time=max_time,
            eval_every_x_epochs=1,
            max_unimproving_models=5,
        )

        return best_model

    def _encoder_targets(self, data):
        """ Encodes target data with a OHE encoder """
        encoded_categories = self.input_encoder.encode(data)
        target = encoded_categories.cpu().numpy()
        target_indexes = np.where(target > 0)[1]
        targets_c = torch.LongTensor(target_indexes)
        labels = targets_c.to(self.net.device)
        return labels

    def _label_targets(self, data):
        """ Encodes target data with a label encoder """
        data = pd.Series(data)
        enc = self.input_encoder.encode(data)
        if len(enc.shape) < 2:
            enc = enc.unsqueeze(-1)
        return enc



import torch
import pandas as pd
import numpy as np
from lightwood.encoder.base import BaseEncoder
from type_infer.dtype import dtype
from lightwood.encoder.helpers import MinMaxNormalizer, CatNormalizer
from lightwood.helpers.general import is_none
from typing import List, Iterable


class ArrayEncoder(BaseEncoder):
    """
    Fits a normalizer for array data.

    To encode, `ArrayEncoder` returns a normalized window of previous data.
    It can be used for generic arrays, as well as for handling historical target values in time series tasks.

    Currently supported normalizing strategies are minmax for numerical arrays, and a simple one-hot for categorical arrays. See `lightwood.encoder.helpers` for more details on each approach.

    """  # noqa

    is_trainable_encoder: bool = True

    def __init__(self, stop_after: float, window: int = None, is_target: bool = False, original_type: dtype = None):
        """
        :param stop_after: time budget in seconds.
        :param window: expected length of array data.
        :param original_type: element-wise data type
        """  # noqa

        super().__init__(is_target)
        self.stop_after = stop_after
        self.original_type = original_type
        self._normalizer = None
        if window is not None:
            self.output_size = window + 1
        else:
            self.output_size = None

    def _pad_and_strip(self, array: List[object]):
        if len(array) < self.output_size:
            array = array + [0] * (self.output_size - len(array))
        if len(array) > self.output_size:
            array = array[:self.output_size]
        return array

    def prepare(self, train_priming_data: Iterable[Iterable], dev_priming_data: Iterable[Iterable]):
        """
        Prepare the array encoder for sequence data.
        :param train_priming_data: Training data of sequences
        :param dev_priming_data: Dev data of sequences
        """
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        priming_data = pd.concat([train_priming_data, dev_priming_data])
        priming_data = priming_data.values

        if self.output_size is None:
            self.output_size = np.max([len(x) for x in priming_data if x is not None])
        for i in range(len(priming_data)):
            if is_none(priming_data[i]):
                priming_data[i] = [0] * self.output_size

        if self.original_type in (dtype.categorical, dtype.binary, dtype.cat_array, dtype.cat_tsarray):
            self._normalizer = CatNormalizer(encoder_class='ordinal')  # maybe turn into OHE encoder?
        else:
            self._normalizer = MinMaxNormalizer()  # maybe turn into numerical encoder?

        if isinstance(priming_data, pd.Series):
            priming_data = priming_data.values

        priming_data = [self._pad_and_strip(list(x)) for x in priming_data]

        self._normalizer.prepare(priming_data)
        self.output_size *= self._normalizer.output_size
        self.is_prepared = True

    def encode(self, column_data: Iterable[Iterable]) -> torch.Tensor:
        """
        Encode the properties of a sequence-of-sequence representation

        :param column_data: Input column data to be encoded
        :returns: a torch-tensor representing the encoded sequence
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if isinstance(column_data, pd.Series):
            column_data = column_data.values

        for i in range(len(column_data)):
            if is_none(column_data[i]):
                column_data[i] = [0] * self.output_size
        column_data = np.array([self._pad_and_strip(list(x)) for x in column_data])

        data = torch.cat([self._normalizer.encode(column_data)], dim=-1)
        data[torch.isnan(data)] = 0.0
        data[torch.isinf(data)] = 0.0

        return data

    def decode(self, data: torch.Tensor) -> List[Iterable]:
        """
        Converts data as a list of arrays.

        :param data: Encoded data prepared by this array encoder
        :returns: A list of iterable sequences in the original data space
        """
        decoded = self._normalizer.decode(data.tolist())
        return decoded


class CatArrayEncoder(ArrayEncoder):
    def __init__(self, stop_after: float, window: int = None, is_target: bool = False):
        super(CatArrayEncoder, self).__init__(stop_after, window, is_target, original_type=dtype.cat_array)

    def prepare(self, train_priming_data: Iterable[Iterable], dev_priming_data: Iterable[Iterable]):
        super().prepare(train_priming_data, dev_priming_data)
        self.index_weights = torch.ones(size=(self.output_size,))

    def decode(self, data: torch.Tensor) -> List[Iterable]:
        data = torch.round(data)  # improves accuracy as by default ordinal encoder will truncate
        decoded = self._normalizer.decode(data.reshape(-1, 1).tolist()).reshape(1, -1)
        return decoded


class NumArrayEncoder(ArrayEncoder):
    def __init__(self, stop_after: float, window: int = None, is_target: bool = False, positive_domain: bool = False):
        self.positive_domain = positive_domain
        super(NumArrayEncoder, self).__init__(stop_after, window, is_target, original_type=dtype.num_array)



import inspect
from typing import List, Tuple, Dict
import torch
import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from lightwood.encoder.base import BaseEncoder


class EncodedDs(Dataset):
    def __init__(self, encoders: Dict[str, BaseEncoder], data_frame: pd.DataFrame, target: str) -> None:
        """
        Create a Lightwood datasource from a data frame and some encoders. This class inherits from `torch.utils.data.Dataset`.
        
        Note: normal behavior is to cache encoded representations to avoid duplicated computations. If you want an option to disable, this please open an issue.
         
        :param encoders: dictionary of Lightwood encoders used to encode the data per each column.
        :param data_frame: original dataframe.
        :param target: name of the target column to predict.
        """  # noqa
        self.data_frame = data_frame
        self.encoders = encoders
        self.target = target
        self.encoder_spans = {}
        self.input_length = 0  # feature tensor dim

        # save encoder span, has to use same iterator as in __getitem__ for correct indeces
        for col in self.data_frame:
            if col != self.target and self.encoders.get(col, False):
                self.encoder_spans[col] = (self.input_length,
                                           self.input_length + self.encoders[col].output_size)
                self.input_length += self.encoders[col].output_size

        # if cache enabled, we immediately build it
        self.use_cache = True
        self.cache_built = False
        self.X_cache: torch.Tensor = torch.full((len(self.data_frame),), fill_value=torch.nan)
        self.Y_cache: torch.Tensor = torch.full((len(self.data_frame),), fill_value=torch.nan)
        self.build_cache()

    def __len__(self):
        """
        The length of an `EncodedDs` datasource equals the amount of rows of the original dataframe.

        :return: length of the `EncodedDs`
        """
        return int(self.data_frame.shape[0])

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        The getter yields a tuple (X, y), where:
          - `X `is a concatenation of all encoded representations of the row. Size: (B, n_features)
          - `y` is the encoded target. Size: (B, n_features)
          
        :param idx: index of the row to access.
        
        :return: tuple (X, y) with encoded data.
        
        """  # noqa
        if self.use_cache and self.X_cache[idx] is not torch.nan:
            X = self.X_cache[idx, :]
            Y = self.Y_cache[idx]
        else:
            X, Y = self._encode_idxs([idx, ])
            if self.use_cache:
                self.X_cache[idx, :] = X
                self.Y_cache[idx, :] = Y

        return X, Y

    def _encode_idxs(self, idxs: list):
        if not isinstance(idxs, list):
            raise Exception(f"Passed indexes is not an iterable. Check the type! Index: {idxs}")

        X = torch.zeros((len(idxs), self.input_length))
        Y = torch.zeros((len(idxs),))
        for col in self.data_frame:
            if self.encoders.get(col, None):
                kwargs = {}
                if 'dependency_data' in inspect.signature(self.encoders[col].encode).parameters:
                    kwargs['dependency_data'] = {dep: [self.data_frame.iloc[idxs][dep]]
                                                 for dep in self.encoders[col].dependencies}
                if hasattr(self.encoders[col], 'data_window'):
                    cols = [self.target] + [f'{self.target}_timestep_{i}'
                                            for i in range(1, self.encoders[col].data_window)]
                    data = self.data_frame[cols].iloc[idxs].values
                else:
                    cols = [col]
                    data = self.data_frame[cols].iloc[idxs].values.flatten()

                encoded_tensor = self.encoders[col].encode(data, **kwargs)
                if torch.isnan(encoded_tensor).any() or torch.isinf(encoded_tensor).any():
                    raise Exception(f'Encoded tensor: {encoded_tensor} contains nan or inf values, this tensor is \
                                      the encoding of column {col} using {self.encoders[col].__class__}')
                if col != self.target:
                    a, b = self.encoder_spans[col]
                    X[:, a:b] = torch.squeeze(encoded_tensor, dim=list(range(2, len(encoded_tensor.shape))))

                # target post-processing
                else:
                    Y = encoded_tensor

                    if len(encoded_tensor.shape) > 2:
                        Y = encoded_tensor.squeeze()

                    if len(encoded_tensor.shape) < 2:
                        Y = encoded_tensor.unsqueeze(1)

                    # else:
                    #     Y = encoded_tensor.ravel()

        return X, Y

    def get_column_original_data(self, column_name: str) -> pd.Series:
        """
        Gets the original data for any given column of the `EncodedDs`.

        :param column_name: name of the column.
        :return:  A `pd.Series` with the original data stored in the `column_name` column.
        """
        return self.data_frame[column_name]

    def get_encoded_column_data(self, column_name: str) -> torch.Tensor:
        """
        Gets the encoded data for any given column of the `EncodedDs`.

        :param column_name: name of the column.
        :return: A `torch.Tensor` with the encoded data of the `column_name` column.
        """
        if self.use_cache and self.cache_built:
            if column_name == self.target and self.Y_cache is not None:
                return self.Y_cache
            elif self.X_cache is not torch.nan:
                a, b = self.encoder_spans[column_name]
                return self.X_cache[:, a:b]

        kwargs = {}
        if 'dependency_data' in inspect.signature(self.encoders[column_name].encode).parameters:
            deps = [dep for dep in self.encoders[column_name].dependencies if dep in self.data_frame.columns]
            kwargs['dependency_data'] = {dep: self.data_frame[dep] for dep in deps}
        encoded_data = self.encoders[column_name].encode(self.data_frame[column_name], **kwargs)
        if torch.isnan(encoded_data).any() or torch.isinf(encoded_data).any():
            raise Exception(f'Encoded tensor: {encoded_data} contains nan or inf values')

        if not isinstance(encoded_data, torch.Tensor):
            raise Exception(
                f'The encoder: {self.encoders[column_name]} for column: {column_name} does not return a Tensor!')

        if self.use_cache and not self.cache_built:
            if column_name == self.target:
                self.Y_cache = encoded_data
            else:
                a, b = self.encoder_spans[column_name]
                self.X_cache = self.X_cache[:, a:b]

        return encoded_data

    def get_encoded_data(self, include_target: bool = True) -> torch.Tensor:
        """
        Gets all encoded data.

        :param include_target: whether to include the target column in the output or not.
        :return: A `torch.Tensor` with the encoded dataframe.
        """
        encoded_dfs = []
        for col in self.data_frame.columns:
            if (include_target or col != self.target) and self.encoders.get(col, False):
                encoded_dfs.append(self.get_encoded_column_data(col))

        return torch.cat(encoded_dfs, 1)

    def build_cache(self):
        """ This method builds a cache for the entire dataframe provided at initialization. """
        if not self.use_cache:
            raise RuntimeError("Cannot build a cache for EncodedDS with `use_cache` set to False.")

        idxs = list(range(len(self.data_frame)))
        X, Y = self._encode_idxs(idxs)
        self.X_cache = X
        self.Y_cache = Y
        self.cache_built = True

    def clear_cache(self):
        """ Clears the `EncodedDs` cache. """
        self.X_cache = torch.full((len(self.data_frame),), fill_value=torch.nan)
        self.Y_cache = torch.full((len(self.data_frame),), fill_value=torch.nan)
        self.cache_built = False


class ConcatedEncodedDs(EncodedDs):
    """
    `ConcatedEncodedDs` abstracts over multiple encoded datasources (`EncodedDs`) as if they were a single entity.
    """  # noqa

    # TODO: We should probably delete this abstraction, it's not really useful and it adds complexity/overhead
    def __init__(self, encoded_ds_arr: List[EncodedDs]) -> None:
        # @TODO: missing super() call here?
        self.encoded_ds_arr = encoded_ds_arr
        self.encoded_ds_lengths = [len(x) for x in self.encoded_ds_arr]
        self.encoders = self.encoded_ds_arr[0].encoders
        self.encoder_spans = self.encoded_ds_arr[0].encoder_spans
        self.target = self.encoded_ds_arr[0].target
        self.data_frame = pd.concat([x.data_frame for x in self.encoded_ds_arr])

    def __len__(self):
        """
        See `lightwood.data.encoded_ds.EncodedDs.__len__()`.
        """
        # @TODO: behavior here is not intuitive
        return max(0, np.sum(self.encoded_ds_lengths) - 2)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        See `lightwood.data.encoded_ds.EncodedDs.__getitem__()`.
        """
        for ds_idx, length in enumerate(self.encoded_ds_lengths):
            if idx - length < 0:
                return self.encoded_ds_arr[ds_idx][idx]
            else:
                idx -= length
        raise StopIteration()

    def get_column_original_data(self, column_name: str) -> pd.Series:
        """
        See `lightwood.data.encoded_ds.EncodedDs.get_column_original_data()`.
        """
        encoded_df_arr = [x.get_column_original_data(column_name) for x in self.encoded_ds_arr]
        return pd.concat(encoded_df_arr)

    def get_encoded_column_data(self, column_name: str) -> torch.Tensor:
        """
        See `lightwood.data.encoded_ds.EncodedDs.get_encoded_column_data()`.
        """
        encoded_df_arr = [x.get_encoded_column_data(column_name) for x in self.encoded_ds_arr]
        return torch.cat(encoded_df_arr, 0)

    def clear_cache(self):
        """
        See `lightwood.data.encoded_ds.EncodedDs.clear_cache()`.
        """
        for ds in self.encoded_ds_arr:
            ds.clear_cache()



import random
from types import SimpleNamespace
from typing import Dict, Optional

import numpy as np
from sklearn.metrics import confusion_matrix

from type_infer.dtype import dtype
from lightwood.analysis.base import BaseAnalysisBlock
from mindsdb_evaluator import evaluate_accuracies


class AccStats(BaseAnalysisBlock):
    """ Computes accuracy stats and a confusion matrix for the validation dataset """

    def __init__(self, deps=('ICP',)):
        super().__init__(deps=deps)
        self.n_decimals = 3

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        ns = SimpleNamespace(**kwargs)

        accuracy_functions = ns.accuracy_functions

        info['score_dict'] = evaluate_accuracies(ns.data, ns.normal_predictions['prediction'],
                                                 ns.target, accuracy_functions, ts_analysis=ns.ts_analysis)

        info['normal_accuracy'] = round(np.mean(list(info['score_dict'].values())), self.n_decimals)
        self.fit(ns, info['result_df'])
        info['val_overall_acc'], info['acc_histogram'], info['cm'], info['acc_samples'] = self.get_accuracy_stats()
        return info

    def fit(self, ns: SimpleNamespace, conf=Optional[np.ndarray]):
        self.col_stats = ns.dtype_dict
        self.target = ns.target
        self.input_cols = ns.input_cols
        self.buckets = ns.stats_info.buckets if ns.stats_info.buckets else {}

        self.normal_predictions_bucketized = []
        self.real_values_bucketized = []
        self.numerical_samples_arr = []

        for n in range(len(ns.normal_predictions)):
            row = ns.data.iloc[n]
            real_value = row[self.target]
            predicted_value = ns.normal_predictions.iloc[n]['prediction']

            if isinstance(predicted_value, list):
                # T+N time series, for now we compare the T+1 prediction only @TODO: generalize
                predicted_value = predicted_value[0]

            predicted_value = predicted_value \
                if self.col_stats[self.target] not in [dtype.integer, dtype.float, dtype.quantity] \
                else float(predicted_value)

            real_value = real_value \
                if self.col_stats[self.target] not in [dtype.integer, dtype.float, dtype.quantity] \
                else float(real_value)

            if self.buckets:
                bucket = self.buckets[self.target]
                predicted_value_b = get_value_bucket(predicted_value, bucket, self.col_stats[self.target])
                real_value_b = get_value_bucket(real_value, bucket, self.col_stats[self.target])
            else:
                predicted_value_b = predicted_value
                real_value_b = real_value

            if conf is not None and self.col_stats[self.target] in [dtype.integer, dtype.float, dtype.quantity]:
                predicted_range = conf.iloc[n][['lower', 'upper']].tolist()
            else:
                predicted_range = (predicted_value_b, predicted_value_b)

            self.real_values_bucketized.append(real_value_b)
            self.normal_predictions_bucketized.append(predicted_value_b)
            if conf is not None and self.col_stats[self.target] in [dtype.integer, dtype.float, dtype.quantity]:
                self.numerical_samples_arr.append((real_value, predicted_range))

    def get_accuracy_stats(self, is_classification=None, is_numerical=None):
        bucket_accuracy = {}
        bucket_acc_counts = {}
        for i, bucket in enumerate(self.normal_predictions_bucketized):
            if bucket not in bucket_acc_counts:
                bucket_acc_counts[bucket] = []

            if len(self.numerical_samples_arr) != 0:
                bucket_acc_counts[bucket].append(self.numerical_samples_arr[i][1][0] <
                                                 self.numerical_samples_arr[i][0] < self.numerical_samples_arr[i][1][1]) # noqa
            else:
                bucket_acc_counts[bucket].append(1 if bucket == self.real_values_bucketized[i] else 0)

        for bucket in bucket_acc_counts:
            bucket_accuracy[bucket] = sum(bucket_acc_counts[bucket]) / len(bucket_acc_counts[bucket])

        accuracy_count = []
        for counts in list(bucket_acc_counts.values()):
            accuracy_count += counts

        overall_accuracy = round(sum(accuracy_count) / len(accuracy_count), self.n_decimals)

        for bucket in range(len(self.buckets)):
            if bucket not in bucket_accuracy:
                if bucket in self.real_values_bucketized:
                    # If it was never predicted, but it did exist as a real value, then assume 0% confidence when it does get predicted # noqa
                    bucket_accuracy[bucket] = 0

        for bucket in range(len(self.buckets)):
            if bucket not in bucket_accuracy:
                # If it wasn't seen either in the real values or in the predicted values, assume average confidence (maybe should be 0 instead ?) # noqa
                bucket_accuracy[bucket] = overall_accuracy

        accuracy_histogram = {
            'buckets': list(bucket_accuracy.keys()),
            'accuracies': list(bucket_accuracy.values()),
            'is_classification': is_classification,
            'is_numerical': is_numerical
        }

        labels = list(set([*self.real_values_bucketized, *self.normal_predictions_bucketized]))
        matrix = confusion_matrix(self.real_values_bucketized, self.normal_predictions_bucketized, labels=labels)
        matrix = [[int(y) if str(y) != 'nan' else 0 for y in x] for x in matrix]

        target_bucket = self.buckets[self.target]
        bucket_values = [target_bucket[i] if i < len(target_bucket) else None for i in labels]

        cm = {
            'matrix': matrix,
            'predicted': bucket_values,
            'real': bucket_values
        }

        accuracy_samples = None
        if len(self.numerical_samples_arr) > 0:
            nr_samples = min(400, len(self.numerical_samples_arr))
            sampled_numerical_samples_arr = random.sample(self.numerical_samples_arr, nr_samples)
            accuracy_samples = {
                'y': [x[0] for x in sampled_numerical_samples_arr],
                'x': [x[1] for x in sampled_numerical_samples_arr]
            }

        return overall_accuracy, accuracy_histogram, cm, accuracy_samples


def get_value_bucket(value, buckets, target_dtype):
    """
    :return: The bucket in the `histogram` in which our `value` falls
    """
    if buckets is None:
        return None

    if target_dtype in (dtype.binary, dtype.categorical):
        if value in buckets:
            bucket = buckets.index(value)
        else:
            bucket = len(buckets)  # for null values

    elif target_dtype in (dtype.integer, dtype.float, dtype.quantity):
        bucket = closest(buckets, value)
    else:
        bucket = len(buckets)  # for null values

    return bucket


def closest(arr, value):
    """
    :return: The index of the member of `arr` which is closest to `value`
    """
    if value is None:
        return -1

    for i, ele in enumerate(arr):
        value = float(str(value).replace(',', '.'))
        if ele > value:
            return i - 1

    return len(arr) - 1



# TODO: add_implicit_values unit test ensures NO changes for a fully specified file.
import inspect

from type_infer.dtype import dtype
from type_infer.base import TypeInformation
from dataprep_ml import StatisticalAnalysis

from lightwood.helpers.templating import _consolidate_analysis_blocks, _add_cls_kwarg
from lightwood.helpers.constants import IMPORTS, IMPORT_EXTERNAL_DIRS
from lightwood.api.types import (
    JsonAI,
    ProblemDefinition,
)
import lightwood.ensemble
import lightwood.encoder


def lookup_encoder(
    col_dtype: str,
    col_name: str,
    is_target: bool,
    problem_defintion: ProblemDefinition,
    is_target_predicting_encoder: bool,
    statistical_analysis: StatisticalAnalysis,
):
    """
    Assign a default encoder for a given column based on its data type, and whether it is a target. Encoders intake raw (but cleaned) data and return an feature representation. This function assigns, per data type, what the featurizer should be. This function runs on each column within the dataset available for model building to assign how it should be featurized.

    Users may override to create a custom encoder to enable their own featurization process. However, in order to generate template JSON-AI, this code runs automatically. Users may edit the generated syntax and use custom approaches while model building.

    For each encoder, "args" may be passed. These args depend an encoder requires during its preparation call.

    :param col_dtype: A data-type of a column specified
    :param col_name: The name of the column
    :param is_target: Whether the column is the target for prediction. If true, only certain possible feature representations are allowed, particularly for complex data types.
    :param problem_definition: The ``ProblemDefinition`` criteria; this populates specifics on how models and encoders may be trained.
    :param is_target_predicting_encoder:
    """  # noqa

    tss = problem_defintion.timeseries_settings
    encoder_lookup = {
        dtype.integer: "NumericEncoder",
        dtype.float: "NumericEncoder",
        dtype.binary: "BinaryEncoder",
        dtype.categorical: "CategoricalAutoEncoder"
        if statistical_analysis is None
        or len(statistical_analysis.histograms[col_name]['x']) > 16
        else "OneHotEncoder",
        dtype.tags: "MultiHotEncoder",
        dtype.date: "DatetimeEncoder",
        dtype.datetime: "DatetimeEncoder",
        dtype.image: "Img2VecEncoder",
        dtype.rich_text: "PretrainedLangEncoder",
        dtype.short_text: "CategoricalAutoEncoder",
        dtype.quantity: "NumericEncoder",
        dtype.audio: "MFCCEncoder",
        dtype.num_array: "NumArrayEncoder",
        dtype.cat_array: "CatArrayEncoder",
        dtype.num_tsarray: "TimeSeriesEncoder",
        dtype.cat_tsarray: "TimeSeriesEncoder",
    }

    # If column is a target, only specific feature representations are allowed that enable supervised tasks
    target_encoder_lookup_override = {
        dtype.rich_text: "VocabularyEncoder",
        dtype.categorical: "OneHotEncoder",
    }

    # Assign a default encoder to each column.
    encoder_dict = {"module": encoder_lookup[col_dtype], "args": {}}

    # If the column is a target, ensure that the feature representation can enable supervised tasks
    if is_target:
        encoder_dict["args"] = {"is_target": "True"}

        if col_dtype in target_encoder_lookup_override:
            encoder_dict["module"] = target_encoder_lookup_override[col_dtype]

        if col_dtype in (dtype.categorical, dtype.binary):
            if problem_defintion.unbias_target:
                encoder_dict["args"][
                    "target_weights"
                ] = "$statistical_analysis.target_weights"
            if problem_defintion.target_weights is not None:
                encoder_dict["args"][
                    "target_weights"
                ] = problem_defintion.target_weights

        if col_dtype in (dtype.integer, dtype.float, dtype.num_array, dtype.num_tsarray):
            encoder_dict["args"][
                "positive_domain"
            ] = "$statistical_analysis.positive_domain"

            if problem_defintion.target_weights is not None:
                encoder_dict["args"][
                    "target_weights"
                ] = problem_defintion.target_weights

    # Time-series representations require more advanced flags
    if tss.is_timeseries:
        gby = tss.group_by if tss.group_by is not None else []

        if tss.order_by in gby:
            raise Exception('The `order_by` column cannot be used to `group_by` simultaneously!')

        if col_name == tss.order_by:
            encoder_dict["module"] = "ArrayEncoder"
            encoder_dict["args"]["original_type"] = f'"{tss.target_type}"'
            encoder_dict["args"]["window"] = f"{tss.window}"

        if is_target:
            if col_dtype in [dtype.integer]:
                encoder_dict["args"]["grouped_by"] = f"{gby}"
                encoder_dict["module"] = "TsNumericEncoder"
            if col_dtype in [dtype.float]:
                encoder_dict["args"]["grouped_by"] = f"{gby}"
                encoder_dict["module"] = "TsNumericEncoder"
            if tss.horizon > 1:
                encoder_dict["args"]["grouped_by"] = f"{gby}"
                encoder_dict["args"]["timesteps"] = f"{tss.horizon}"
                if col_dtype in [dtype.num_tsarray]:
                    encoder_dict["module"] = "TsArrayNumericEncoder"
                elif col_dtype in [dtype.cat_tsarray]:
                    encoder_dict["module"] = "TsCatArrayEncoder"

        if "__mdb_ts_previous" in col_name or col_name in tss.historical_columns:
            encoder_dict["module"] = "TimeSeriesEncoder"
            encoder_dict["args"]["original_type"] = f'"{tss.target_type}"'
            encoder_dict["args"]["window"] = f"{tss.window}"

    # Set arguments for the encoder
    if encoder_dict["module"] == "PretrainedLangEncoder" and not is_target:
        encoder_dict["args"]["output_type"] = "$dtype_dict[$target]"

    if eval(encoder_dict["module"]).is_trainable_encoder:
        encoder_dict["args"]["stop_after"] = "$problem_definition.seconds_per_encoder"

    if is_target_predicting_encoder:
        encoder_dict["args"]["embed_mode"] = "False"
    return encoder_dict


def generate_json_ai(
    type_information: TypeInformation,
    statistical_analysis: StatisticalAnalysis,
    problem_definition: ProblemDefinition,
) -> JsonAI:
    """
    Given ``type_infer.TypeInformation``, ``dataprep_ml.StatisticalAnalysis``, and the ``ProblemDefinition``, generate a JSON config file with the necessary elements of the ML pipeline populated.

    :param TypeInformation: Specifies what data types each column within the dataset are. Generated by `mindsdb/type_infer`.
    :param statistical_analysis:
    :param problem_definition: Specifies details of the model training/building procedure, as defined by ``ProblemDefinition``

    :returns: JSON-AI object with fully populated details of the ML pipeline
    """  # noqaexec
    exec(IMPORTS, globals())
    exec(IMPORT_EXTERNAL_DIRS, globals())
    target = problem_definition.target
    input_cols = []
    dependency_dict = {}
    tss = problem_definition.timeseries_settings

    dtype_dict_override = problem_definition.dtype_dict
    dtype_dict = type_information.dtypes

    for k in type_information.identifiers:
        if not (tss.is_timeseries and tss.group_by and k in tss.group_by) and k != target:
            del dtype_dict[k]

    for k, v in dtype_dict_override.items():
        dtype_dict[k] = v

    for col_name, col_dtype in dtype_dict.items():
        if (
                (col_name not in type_information.identifiers
                 and col_dtype not in (dtype.invalid, dtype.empty)
                 and col_name != target)
                or
                (tss.group_by is not None and col_name in tss.group_by)
        ):
            if col_name != problem_definition.target:
                input_cols.append(col_name)

    is_target_predicting_encoder = False
    is_ts = problem_definition.timeseries_settings.is_timeseries
    imputers = []

    # Single text column classification
    if (
        len(input_cols) == 1
        and type_information.dtypes[input_cols[0]] in (dtype.rich_text)
        and type_information.dtypes[target] in (dtype.categorical, dtype.binary)
    ):
        is_target_predicting_encoder = True

    submodels = []
    if is_target_predicting_encoder:
        submodels.extend(
            [
                {
                    "module": "Unit",
                    "args": {
                        "target_encoder": "$encoders[self.target]",
                        "stop_after": "$problem_definition.seconds_per_mixer",
                    },
                }
            ]
        )
    else:

        # add neural model
        if not tss.is_timeseries:
            submodels.extend(
                [
                    {
                        "module": "Neural",
                        "args": {
                            "fit_on_dev": True,
                            "stop_after": "$problem_definition.seconds_per_mixer",
                            "search_hyperparameters": True,
                        },
                    }
                ]
            )
        else:
            submodels.extend(
                [
                    {
                        "module": "NeuralTs",
                        "args": {
                            "fit_on_dev": True,
                            "stop_after": "$problem_definition.seconds_per_mixer",
                            "search_hyperparameters": True,
                        },
                    },
                ]
            )

        # add other models
        if (not tss.is_timeseries or tss.horizon == 1) and dtype_dict[target] not in (dtype.num_array, dtype.cat_array):
            submodels.extend(
                [
                    {
                        "module": "XGBoostMixer",
                        "args": {
                            "stop_after": "$problem_definition.seconds_per_mixer",
                            "fit_on_dev": True,
                        },
                    },
                    {
                        "module": "Regression",
                        "args": {
                            "stop_after": "$problem_definition.seconds_per_mixer",
                        },
                    },
                    {
                        "module": "RandomForest",
                        "args": {
                            "stop_after": "$problem_definition.seconds_per_mixer",
                            "fit_on_dev": True,
                        },
                    },
                ]
            )

        # special forecasting dispatch
        elif tss.is_timeseries:
            submodels.extend([
                {
                    "module": "XGBoostArrayMixer",
                    "args": {},
                },
            ])

    model = {
        "module": "BestOf",
        "args": {
            "submodels": submodels,
        }
    }

    num_ts_dtypes = (dtype.integer, dtype.float, dtype.quantity)
    if tss.is_timeseries and tss.horizon > 1:
        if dtype_dict[target] in num_ts_dtypes:
            dtype_dict[target] = dtype.num_tsarray
            problem_definition.anomaly_detection = True
        else:
            dtype_dict[target] = dtype.cat_tsarray
    elif tss.is_timeseries and dtype_dict[target] in num_ts_dtypes:
        problem_definition.anomaly_detection = True

    encoders = {
        target: lookup_encoder(
            dtype_dict[target],
            target,
            True,
            problem_definition,
            False,
            statistical_analysis,
        )
    }

    for col in input_cols:
        encoders[col] = lookup_encoder(
            dtype_dict[col],
            col,
            False,
            problem_definition,
            is_target_predicting_encoder,
            statistical_analysis,
        )

    # Decide on the accuracy functions to use
    output_dtype = dtype_dict[target]
    if output_dtype in [
        dtype.integer,
        dtype.float,
        dtype.date,
        dtype.datetime,
        dtype.quantity,
    ]:
        accuracy_functions = ["r2_score"]
    elif output_dtype in [dtype.categorical, dtype.tags, dtype.binary]:
        accuracy_functions = ["balanced_accuracy_score"]
    elif output_dtype in (dtype.num_tsarray, ):
        accuracy_functions = ["complementary_smape_array_accuracy"]
    elif output_dtype in (dtype.num_array, ):
        accuracy_functions = ["evaluate_num_array_accuracy"]
    elif output_dtype in (dtype.cat_array, dtype.cat_tsarray):
        accuracy_functions = ["evaluate_cat_array_accuracy"]
    else:
        raise Exception(
            f"Please specify a custom accuracy function for output type {output_dtype}"
        )

    if is_ts:
        if output_dtype in [dtype.integer, dtype.float, dtype.quantity]:
            # forces this acc fn for t+1 time series forecasters
            accuracy_functions = ["complementary_smape_array_accuracy"]

        if output_dtype in (dtype.integer, dtype.float, dtype.quantity, dtype.num_tsarray):
            imputers.append({"module": "NumericalImputer",
                             "args": {
                                 "value": "'zero'",
                                 "target": f"'{target}'"}}
                            )
        elif output_dtype in [dtype.categorical, dtype.tags, dtype.binary, dtype.cat_tsarray]:
            imputers.append({"module": "CategoricalImputer",
                             "args": {
                                 "value": "'mode'",
                                 "target": f"'{target}'"}}
                            )

    if problem_definition.time_aim is None:
        # 5 days
        problem_definition.time_aim = 3 * 24 * 3600

    # Encoders are assigned 1/3 of the time unless a user overrides this (equal time per encoder)
    if problem_definition.seconds_per_encoder is None:
        nr_trainable_encoders = len(
            [
                x
                for x in encoders.values()
                if eval(x["module"]).is_trainable_encoder
            ]
        )
        if nr_trainable_encoders > 0:
            problem_definition.seconds_per_encoder = 0.33 * problem_definition.time_aim / nr_trainable_encoders

    # Mixers are assigned 1/3 of the time aim (or 2/3 if there are no trainable encoders )\
    # unless a user overrides this (equal time per mixer)
    if problem_definition.seconds_per_mixer is None:
        if problem_definition.seconds_per_encoder is None:
            problem_definition.seconds_per_mixer = 0.66 * problem_definition.time_aim / len(model['args']['submodels'])
        else:
            problem_definition.seconds_per_mixer = 0.33 * problem_definition.time_aim / len(model['args']['submodels'])

    return JsonAI(
        cleaner=None,
        splitter=None,
        analyzer=None,
        explainer=None,
        encoders=encoders,
        imputers=imputers,
        dtype_dict=dtype_dict,
        dependency_dict=dependency_dict,
        model=model,
        problem_definition=problem_definition,
        identifiers=type_information.identifiers,
        timeseries_transformer=None,
        timeseries_analyzer=None,
        accuracy_functions=accuracy_functions,
    )


def _merge_implicit_values(field: dict, implicit_value: dict) -> dict:
    """
    Helper function for `_populate_implicit_field`.
    Takes a user-defined field along with its implicit value, and merges them together.

    :param field: JsonAI field with user-defined parameters.
    :param implicit_value: implicit values for the field.
    :return: original field with implicit values merged into it.
    """
    exec(IMPORTS, globals())
    exec(IMPORT_EXTERNAL_DIRS, globals())
    module = eval(field["module"])

    if inspect.isclass(module):
        args = list(inspect.signature(module.__init__).parameters.keys())[1:]
    else:
        args = module.__code__.co_varnames

    for arg in args:
        if "args" not in field:
            field["args"] = implicit_value["args"]
        else:
            if arg not in field["args"]:
                if arg in implicit_value["args"]:
                    field["args"][arg] = implicit_value["args"][arg]

    return field


def _populate_implicit_field(
    json_ai: JsonAI, field_name: str, implicit_value: dict, is_timeseries: bool
) -> None:
    """
    Populate the implicit field of the JsonAI, either by filling it in entirely if missing, or by introspecting the class or function and assigning default values to the args in it's signature that are in the implicit default but haven't been populated by the user

    :params: json_ai: ``JsonAI`` object that describes the ML pipeline that may not have every detail fully specified.
    :params: field_name: Name of the field the implicit field in ``JsonAI``
    :params: implicit_value: The dictionary containing implicit values for the module and arg in the field
    :params: is_timeseries: Whether or not this is a timeseries problem

    :returns: nothing, this method mutates the respective field of the ``JsonAI`` object it receives
    """  # noqa
    # These imports might be slow, in which case the only <easy> solution is to line this code
    field = json_ai.__getattribute__(field_name)
    if field is None:
        # This if is to only populated timeseries-specific implicit fields for implicit problems
        if is_timeseries or field_name not in (
            "timeseries_analyzer",
            "timeseries_transformer",
        ):
            field = implicit_value

    # If the user specified one or more subfields in a field that's a list
    # Populate them with implicit arguments form the implicit values from that subfield
    elif isinstance(field, list) and isinstance(implicit_value, list):
        for i in range(len(field)):
            sub_field_implicit = [
                x for x in implicit_value if x["module"] == field[i]["module"]
            ]
            if len(sub_field_implicit) == 1:
                field[i] = _merge_implicit_values(field[i], sub_field_implicit[0])
        for sub_field_implicit in implicit_value:
            if (
                len([x for x in field if x["module"] == sub_field_implicit["module"]])
                == 0
            ):
                field.append(sub_field_implicit)
    # If the user specified the field, add implicit arguments which we didn't specify
    else:
        field = _merge_implicit_values(field, implicit_value)
    json_ai.__setattr__(field_name, field)


def add_implicit_values(json_ai: JsonAI) -> JsonAI:
    """
    To enable brevity in writing, auto-generate the "unspecified/missing" details required in the ML pipeline.

    :params: json_ai: ``JsonAI`` object that describes the ML pipeline that may not have every detail fully specified.

    :returns: ``JSONAI`` object with all necessary parameters that were previously left unmentioned filled in.
    """
    problem_definition = json_ai.problem_definition
    tss = problem_definition.timeseries_settings
    is_ts = tss.is_timeseries
    mixers = json_ai.model['args']['submodels']

    # Add implicit ensemble arguments
    param_pairs = {
        'target': json_ai.model["args"].get("target", "$target"),
        'data': json_ai.model["args"].get("data", "encoded_test_data"),
        'mixers': json_ai.model["args"].get("mixers", "$mixers"),
        'fit': json_ai.model["args"].get("fit", True),
        'args': json_ai.model["args"].get("args", "$pred_args"),
        'accuracy_functions': json_ai.model["args"].get("accuracy_functions", "$accuracy_functions"),
        'ts_analysis': json_ai.model["args"].get("ts_analysis", "self.ts_analysis" if is_ts else None),
        'dtype_dict': json_ai.model["args"].get("dtype_dict", "$dtype_dict"),
    }
    ensemble_cls = getattr(lightwood.ensemble, json_ai.model["module"])
    filtered_params = {}
    for p_name, p_value in param_pairs.items():
        _add_cls_kwarg(ensemble_cls, filtered_params, p_name, p_value)

    json_ai.model["args"] = filtered_params
    json_ai.model["args"]['submodels'] = mixers  # add mixers back in

    # Add implicit mixer arguments
    for i in range(len(mixers)):
        if not mixers[i].get("args", False):
            mixers[i]["args"] = {}

        if mixers[i]["module"] == "Unit":
            continue

        # common
        mixers[i]["args"]["target"] = mixers[i]["args"].get("target", "$target")
        mixers[i]["args"]["dtype_dict"] = mixers[i]["args"].get("dtype_dict", "$dtype_dict")
        mixers[i]["args"]["stop_after"] = mixers[i]["args"].get("stop_after", "$problem_definition.seconds_per_mixer")

        # specific
        if mixers[i]["module"] in ("Neural", "NeuralTs", "TabTransformerMixer"):
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )

            if mixers[i]["module"] in ("Neural", "NeuralTs"):
                mixers[i]["args"]["net"] = mixers[i]["args"].get(
                    "net",
                    '"DefaultNet"'
                    if not tss.is_timeseries or not tss.use_previous_target
                    else '"ArNet"',
                )
                mixers[i]["args"]["search_hyperparameters"] = mixers[i]["args"].get("search_hyperparameters", True)
                mixers[i]["args"]["fit_on_dev"] = mixers[i]["args"].get("fit_on_dev", True)

            if mixers[i]["module"] == "NeuralTs":
                mixers[i]["args"]["timeseries_settings"] = mixers[i]["args"].get(
                    "timeseries_settings", "$problem_definition.timeseries_settings"
                )
                mixers[i]["args"]["ts_analysis"] = mixers[i]["args"].get("ts_analysis", "$ts_analysis")

            if mixers[i]["module"] == "TabTransformerMixer":
                mixers[i]["args"]["search_hyperparameters"] = mixers[i]["args"].get("search_hyperparameters", False)
                mixers[i]["args"]["fit_on_dev"] = mixers[i]["args"].get("fit_on_dev", False)

        elif mixers[i]["module"] in ("LightGBM", "XGBoostMixer"):
            mixers[i]["args"]["input_cols"] = mixers[i]["args"].get(
                "input_cols", "$input_cols"
            )
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )
            mixers[i]["args"]["fit_on_dev"] = mixers[i]["args"].get(
                "fit_on_dev", True
            )
            mixers[i]["args"]["use_optuna"] = True

        elif mixers[i]["module"] == "Regression":
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )

        elif mixers[i]["module"] == "RandomForest":
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )

        elif mixers[i]["module"] in ("LightGBMArray", "XGBoostArrayMixer"):
            mixers[i]["args"]["input_cols"] = mixers[i]["args"].get(
                "input_cols", "$input_cols"
            )
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )
            mixers[i]["args"]["tss"] = mixers[i]["args"].get("tss", "$problem_definition.timeseries_settings")
            mixers[i]["args"]["ts_analysis"] = mixers[i]["args"].get("ts_analysis", "$ts_analysis")
            mixers[i]["args"]["fit_on_dev"] = mixers[i]["args"].get("fit_on_dev", "True")
            mixers[i]["args"]["use_stl"] = mixers[i]["args"].get("use_stl", "False")

        elif mixers[i]["module"] in ("NHitsMixer", "GluonTSMixer"):
            mixers[i]["args"]["horizon"] = "$problem_definition.timeseries_settings.horizon"
            mixers[i]["args"]["window"] = "$problem_definition.timeseries_settings.window"
            mixers[i]["args"]["ts_analysis"] = mixers[i]["args"].get(
                "ts_analysis", "$ts_analysis"
            )
            problem_definition.fit_on_all = False  # takes too long otherwise

        elif mixers[i]["module"] in ("SkTime", "ProphetMixer", "ETSMixer", "ARIMAMixer"):
            mixers[i]["args"]["ts_analysis"] = mixers[i]["args"].get(
                "ts_analysis", "$ts_analysis"
            )
            if "horizon" not in mixers[i]["args"]:
                mixers[i]["args"]["horizon"] = "$problem_definition.timeseries_settings.horizon"

            # enforce fit_on_all if this mixer is specified
            problem_definition.fit_on_all = True

    # encoder checks
    for name in json_ai.encoders:
        if name not in json_ai.dependency_dict:
            json_ai.dependency_dict[name] = []

    # filter arguments for included encoders (custom encoders will skip the check)
    for col, enc_dict in json_ai.encoders.items():
        filtered_kwargs = {}
        if hasattr(lightwood.encoder, enc_dict['module']):
            encoder_cls = getattr(lightwood.encoder, enc_dict['module'])
            for k, v in enc_dict['args'].items():
                _add_cls_kwarg(encoder_cls, filtered_kwargs, k, v)
            json_ai.encoders[col]['args'] = filtered_kwargs

    # Add "hidden" fields
    hidden_fields = {
        "cleaner": {
            "module": "cleaner",
            "args": {
                "pct_invalid": "$problem_definition.pct_invalid",
                "identifiers": "$identifiers",
                "data": "data",
                "dtype_dict": "$dtype_dict",
                "target": "$target",
                "mode": "$mode",
                "imputers": "$imputers",
                "timeseries_settings": "$problem_definition.timeseries_settings.to_dict()",
                "anomaly_detection": "$problem_definition.anomaly_detection",
            },
        },
        "splitter": {
            "module": "splitter",
            "args": {
                "tss": "$problem_definition.timeseries_settings.to_dict()",
                "data": "data",
                "seed": "$problem_definition.seed_nr",
                "target": "$target",
                "dtype_dict": "$dtype_dict",
                "pct_train": 0.8,
                "pct_dev": 0.1,
                "pct_test": 0.1,
            },
        },
        "analyzer": {
            "module": "model_analyzer",
            "args": {
                "stats_info": "$statistical_analysis",
                "pdef": "$problem_definition",
                "accuracy_functions": "$accuracy_functions",
                "predictor": "$ensemble",
                "data": "encoded_test_data",
                "train_data": "encoded_train_data",
                "target": "$target",
                "dtype_dict": "$dtype_dict",
                "analysis_blocks": "$analysis_blocks",
                "ts_analysis": "$ts_analysis" if is_ts else None,
            },
        },
        "explainer": {
            "module": "explain",
            "args": {
                "problem_definition": "$problem_definition",
                "stat_analysis": "$statistical_analysis",
                "data": "data",
                "encoded_data": "encoded_data",
                "predictions": "df",
                "runtime_analysis": "$runtime_analyzer",
                "ts_analysis": "$ts_analysis" if is_ts else None,
                "target_name": "$target",
                "target_dtype": "$dtype_dict[self.target]",
                "explainer_blocks": "$analysis_blocks",
                "pred_args": "$pred_args",
            },
        },
        "analysis_blocks": [
            {
                "module": "ICP",
                "args": {
                    "fixed_significance": None,
                    "confidence_normalizer": False,
                },
            },
            {
                "module": "AccStats",
                "args": {"deps": ["ICP"]},
            },
            {
                "module": "ConfStats",
                "args": {"deps": ["ICP"]},
            },
            {
                "module": "PermutationFeatureImportance",
                "args": {"deps": ["AccStats"]},
            },
        ] if problem_definition.use_default_analysis else [],
        "timeseries_transformer": {
            "module": "transform_timeseries",
            "args": {
                "timeseries_settings": "$problem_definition.timeseries_settings",
                "data": "data",
                "dtype_dict": "$dtype_dict",
                "target": "$target",
                "mode": "$mode",
                "pred_args": "$pred_args",
            },
        },
        "timeseries_analyzer": {
            "module": "timeseries_analyzer",
            "args": {
                "timeseries_settings": "$problem_definition.timeseries_settings",
                "data": "data",
                "dtype_dict": "$dtype_dict",
                "target": "$target",
            },
        },
    }

    for field_name, implicit_value in hidden_fields.items():
        _populate_implicit_field(json_ai, field_name, implicit_value, tss.is_timeseries)

    # further consolidation
    to_inspect = ['analysis_blocks']
    consolidation_methods = {
        'analysis_blocks': _consolidate_analysis_blocks
    }
    for k in to_inspect:
        method = consolidation_methods[k]
        setattr(json_ai, k, method(json_ai, k))

    return json_ai


def validate_json_ai(json_ai: JsonAI) -> bool:
    """
    Checks the validity of a ``JsonAI`` object

    :param json_ai: A ``JsonAI`` object

    :returns: Whether the JsonAI is valid, i.e. doesn't contain prohibited values, unknown values and can be turned into code.
    """ # noqa
    from lightwood.api.high_level import predictor_from_code, code_from_json_ai

    try:
        predictor_from_code(code_from_json_ai(json_ai))
        return True
    except Exception:
        return False



from copy import deepcopy
from types import SimpleNamespace
from typing import Dict

import numpy as np
from sklearn.utils import shuffle

from lightwood.helpers.log import log
from lightwood.data.encoded_ds import EncodedDs
from lightwood.analysis.base import BaseAnalysisBlock
from mindsdb_evaluator import evaluate_accuracies
from lightwood.api.types import PredictionArguments


class PermutationFeatureImportance(BaseAnalysisBlock):
    """
    Analysis block that estimates column importances via permutation.

    Roughly speaking, the procedure:
        - iterates over all input columns
        - if the input column is optional, shuffle its values, then generate predictions for the input data
        - compare this accuracy with the accuracy obtained using unshuffled data
        - all accuracy differences are normalized with respect to the original accuracy (clipped at zero if negative)
        - report these as estimated column importance scores

    Note that, crucially, this method does not refit the predictor at any point.

    :param row_limit: Set to 0 to use the entire validation dataset.
    :param col_limit: Set to 0 to consider all possible columns.

    Reference:
        https://scikit-learn.org/stable/modules/permutation_importance.html
        https://compstat-lmu.github.io/iml_methods_limitations/pfi.html
    """
    def __init__(self, disable_column_importance=False, row_limit=1000, col_limit=10, deps=tuple('AccStats',)):
        super().__init__(deps=deps)
        self.disable_column_importance = disable_column_importance
        self.row_limit = row_limit
        self.col_limit = col_limit
        self.n_decimals = 3

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        ns = SimpleNamespace(**kwargs)

        if self.disable_column_importance:
            info['column_importances'] = None
        elif ns.tss.is_timeseries or ns.has_pretrained_text_enc:
            log.warning(f"Block 'PermutationFeatureImportance' does not support time series nor text encoding, skipping...")  # noqa
            info['column_importances'] = None
        else:
            if self.row_limit:
                log.info(f"[PFI] Using a random sample ({self.row_limit} rows out of {len(ns.encoded_val_data.data_frame)}).")  # noqa
                ref_df = ns.encoded_val_data.data_frame.sample(frac=1).reset_index(drop=True).iloc[:self.row_limit]
            else:
                log.info(f"[PFI] Using complete validation set ({len(ns.encoded_val_data.data_frame)} rows).")
                ref_df = deepcopy(ns.encoded_val_data.data_frame)

            ref_data = EncodedDs(ns.encoded_val_data.encoders, ref_df, ns.target)

            args = {'predict_proba': True} if ns.is_classification else {}
            ref_preds = ns.predictor(ref_data, args=PredictionArguments.from_dict(args))
            ref_score = np.mean(list(evaluate_accuracies(ref_data.data_frame,
                                                         ref_preds['prediction'],
                                                         ns.target,
                                                         ns.accuracy_functions
                                                         ).values()))
            shuffled_col_accuracy = {}
            shuffled_cols = []
            for x in ns.input_cols:
                if ('__mdb' not in x) and \
                        (not ns.tss.is_timeseries or (x != ns.tss.order_by and x not in ns.tss.historical_columns)):
                    shuffled_cols.append(x)

            if self.col_limit:
                shuffled_cols = shuffled_cols[:min(self.col_limit, len(ns.encoded_val_data.data_frame.columns))]
                log.info(f"[PFI] Set to consider first {self.col_limit} columns out of {len(shuffled_cols)}: {shuffled_cols}.")  # noqa
            else:
                log.info(f"[PFI] Computing importance for all {len(shuffled_cols)} columns: {shuffled_cols}")

            for col in shuffled_cols:
                shuffle_data = deepcopy(ref_data)
                shuffle_data.clear_cache()
                shuffle_data.data_frame[col] = shuffle(shuffle_data.data_frame[col].values)
                shuffle_data.build_cache()  # TODO: bottleneck, add a method to build a single column instead!

                shuffled_preds = ns.predictor(shuffle_data, args=PredictionArguments.from_dict(args))
                shuffled_col_accuracy[col] = np.mean(list(evaluate_accuracies(
                    shuffle_data.data_frame,
                    shuffled_preds['prediction'],
                    ns.target,
                    ns.accuracy_functions
                ).values()))

            column_importances = {}
            acc_increases = np.zeros((len(shuffled_cols),))
            for i, col in enumerate(shuffled_cols):
                accuracy_increase = (ref_score - shuffled_col_accuracy[col])
                acc_increases[i] = round(accuracy_increase, self.n_decimals)
            for col, inc in zip(shuffled_cols, acc_increases):
                column_importances[col] = inc

            info['column_importances'] = column_importances

        return info




import pandas as pd

from lightwood.api.types import TimeseriesSettings
from type_infer.dtype import dtype
from lightwood.encoder.helpers import MinMaxNormalizer, CatNormalizer


def generate_target_group_normalizers(
        data: pd.DataFrame,
        target: str,
        dtype_dict: dict,
        tss: TimeseriesSettings
):
    """
    Helper function called from data_source. It generates and fits all needed normalizers for a target variable based on its grouped entities.
    
    :return: modified data with dictionary with normalizers for said target variable based on some grouped-by columns
    """  # noqa
    normalizers = {}
    target_dtype = dtype_dict[target]

    # categorical normalizers
    if target_dtype in (dtype.categorical, dtype.binary, dtype.cat_tsarray):
        normalizers['__default'] = CatNormalizer()
        normalizers['__default'].prepare(data)

    # numerical normalizers, here we spawn one per each group combination
    else:
        grouped = data.groupby(by=tss.group_by) if tss.group_by else data.groupby(lambda x: '__default')
        for name, subset in grouped:
            if subset.shape[0] > 0:
                target_data = subset[target].values
                if target_dtype == dtype.num_tsarray:
                    target_data = target_data.reshape(-1, 1).astype(float)

                normalizers[name] = MinMaxNormalizer(combination=name)
                normalizers[name].prepare(target_data)

        if not normalizers.get('__default'):
            # ...plus a default one, used at inference time and fitted with all training data
            normalizers['__default'] = MinMaxNormalizer()
            normalizers['__default'].prepare(data[target].values)

    return normalizers



from lightwood.data.timeseries_transform import transform_timeseries
from lightwood.data.timeseries_analyzer import timeseries_analyzer
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs

__all__ = ['transform_timeseries', 'timeseries_analyzer', 'EncodedDs', 'ConcatedEncodedDs']



from lightwood.encoder.text.pretrained import PretrainedLangEncoder
from lightwood.encoder.text.tfidf import TfidfEncoder
from lightwood.encoder.text.short import ShortTextEncoder
from lightwood.encoder.text.vocab import VocabularyEncoder


__all__ = ['PretrainedLangEncoder', 'TfidfEncoder', 'ShortTextEncoder', 'VocabularyEncoder']



import math
import torch
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.helpers.device import get_device_from_name
from lightwood.helpers.log import log
import numpy as np


class DefaultNet(torch.nn.Module):
    """
    Default neural network module used in `Neural` mixer to learn the predictive task based on encoded feature representations.
    
    Small architecture with either a one or two linear layers. Hidden size might be constrained to have at most `max_params` in the network, trading off accuracy for speed.
    """  # noqa

    def __init__(self,
                 input_size: int = None,
                 output_size: int = None,
                 shape: list = None,
                 max_params: int = int(3e7),
                 num_hidden: int = 1,
                 dropout: float = 0,
                 device: str = '') -> None:

        super(DefaultNet, self).__init__()

        if input_size is not None and output_size is not None:
            self.input_size = input_size
            self.output_size = output_size
            hidden_size = max([self.input_size * 2, self.output_size * 2, 400])
            shape = [self.input_size] + [hidden_size] * num_hidden + [self.output_size]

            # If the network is too big, shrink it
            if np.sum([shape[i] * shape[i + 1] for i in range(len(shape) - 1)]) > max_params:
                log.warning('Shrinking network!')
                hidden_size = math.floor(max_params / (self.input_size * self.output_size))

                if hidden_size > self.output_size:
                    shape = [self.input_size, hidden_size, self.output_size]
                else:
                    shape = [self.input_size, self.output_size]
        if shape is not None:
            layers = []
            for ind in range(len(shape) - 1):
                layers.append(torch.nn.Linear(shape[ind], shape[ind + 1]))
                if ind < len(shape) - 2:
                    layers.append(torch.nn.SELU())
                    if dropout > 0.001:
                        layers.append(torch.nn.Dropout(p=dropout))
        else:
            raise Exception('You must specify other a shape or an input and output size when creating a DefaultNet!')

        self.net = torch.nn.Sequential(*layers)

        self.to(get_device_from_name(device))

    def to(self, device: torch.device) -> torch.nn.Module:
        if 'cuda' not in str(torch.device) == 0:
            log.warning(
                'Creating neural network on CPU, it will be significantly slower than using a GPU, consider using a GPU instead') # noqa
        self.net = self.net.to(device)

        self.device = device
        return self

    def forward(self, input):
        try:
            with LightwoodAutocast():
                output = self.net(input)
        except Exception:
            output = self.net(input)
        return output



import time
from copy import deepcopy
from collections import deque
from typing import Dict, List, Optional

import torch
import numpy as np
import pandas as pd
from torch import nn
import torch_optimizer as ad_optim
from sklearn.metrics import r2_score
from torch.cuda.amp import GradScaler
from torch.utils.data import DataLoader
from torch.nn.modules.loss import MSELoss
from torch.optim.optimizer import Optimizer

from type_infer.dtype import dtype
from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.data.encoded_ds import EncodedDs
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.helpers.ar_net import ArNet
from lightwood.mixer.helpers.default_net import DefaultNet
from lightwood.api.types import PredictionArguments
from lightwood.mixer.helpers.transform_corss_entropy_loss import TransformCrossEntropyLoss


class Neural(BaseMixer):
    model: nn.Module
    dtype_dict: dict
    target: str
    epochs_to_best: int
    fit_on_dev: bool
    supports_proba: bool

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            target_encoder: BaseEncoder,
            net: str,
            fit_on_dev: bool,
            search_hyperparameters: bool,
            n_epochs: Optional[int] = None,
            lr: Optional[float] = None,
    ):
        """
        The Neural mixer trains a fully connected dense network from concatenated encoded outputs of each of the features in the dataset to predicted the encoded output. 
        
        :param stop_after: How long the total fitting process should take
        :param target: Name of the target column
        :param dtype_dict: Data type dictionary
        :param target_encoder: Reference to the encoder used for the target
        :param net: The network type to use (`DeafultNet` or `ArNet`)
        :param fit_on_dev: If we should fit on the dev dataset
        :param search_hyperparameters: If the network should run a more through hyperparameter search (currently disabled)
        :param n_epochs: amount of epochs that the network will be trained for. Supersedes all other early stopping criteria if specified.
        :param lr: learning rate for the network. By default, it is automatically selected based on an initial search process.
        """ # noqa
        super().__init__(stop_after)
        self.dtype_dict = dtype_dict
        self.target = target
        self.target_encoder = target_encoder
        self.num_hidden = 1
        self.epochs_to_best = 0
        self.n_epochs = n_epochs
        self.lr = lr
        self.loss_hist_len = 7  # length of queue to use for early stopping
        self.fit_on_dev = fit_on_dev
        self.net_name = net
        self.supports_proba = dtype_dict[target] in [dtype.binary, dtype.categorical]
        self.search_hyperparameters = search_hyperparameters
        self.stable = True

    def _final_tuning(self, data):
        if self.dtype_dict[self.target] in (dtype.integer, dtype.float, dtype.quantity):
            self.model = self.model.eval()
            with torch.no_grad():
                acc_dict = {}
                for decode_log in [True, False]:
                    self.target_encoder.decode_log = decode_log
                    decoded_predictions = []
                    decoded_real_values = []
                    for X, Y in data:
                        X = X.to(self.model.device)
                        Y = Y.to(self.model.device)
                        Yh = self._net_call(X)

                        Yh = torch.unsqueeze(Yh, 0) if len(Yh.shape) < 2 else Yh
                        Y = torch.unsqueeze(Y, 0) if len(Y.shape) < 2 else Y

                        decoded_predictions.extend(self.target_encoder.decode(Yh))
                        decoded_real_values.extend(self.target_encoder.decode(Y))

                    acc_dict[decode_log] = r2_score(decoded_real_values, decoded_predictions)

            self.target_encoder.decode_log = acc_dict[True] > acc_dict[False]

    def _select_criterion(self) -> torch.nn.Module:
        if self.dtype_dict[self.target] in (dtype.categorical, dtype.binary):
            criterion = TransformCrossEntropyLoss(weight=self.target_encoder.index_weights.to(self.model.device))
        elif self.dtype_dict[self.target] in (dtype.tags, dtype.cat_tsarray):
            criterion = nn.BCEWithLogitsLoss()
        elif self.dtype_dict[self.target] in (dtype.cat_array, ):
            criterion = nn.L1Loss()
        elif self.dtype_dict[self.target] in (dtype.integer, dtype.float, dtype.quantity, dtype.num_array):
            criterion = MSELoss()
        else:
            criterion = MSELoss()

        return criterion

    def _select_optimizer(self, model, lr) -> Optimizer:
        optimizer = ad_optim.Ranger(model.parameters(), lr=lr, weight_decay=2e-2)
        return optimizer

    def _find_lr(self, train_data):
        lr = 1e-4  # good starting point as search escalates
        lrs = deque([5e-4, 1e-3, 2e-3, 3e-3, 5e-3, 1e-2, 5e-2, 1e-1])
        starting_model = deepcopy(self.model)
        criterion = self._select_criterion()
        scaler = GradScaler()

        running_losses = deque(maxlen=self.loss_hist_len)
        lr_log = deque(maxlen=self.loss_hist_len)
        best_model = self.model
        stop = False

        n_steps = 10
        cum_loss = 0

        while stop is False:
            # overfit learning on first n_steps samples (biased, but we only want an intuition on what LR is decent)
            dl = DataLoader(train_data,
                            batch_size=min(len(train_data.data_frame), 32, self.batch_size),
                            shuffle=False)
            dl_iter = iter(dl)
            self.model = deepcopy(starting_model)
            self.model.train()
            optimizer = self._select_optimizer(self.model, lr=lr)

            for i in range(n_steps):
                try:
                    X, Y = next(dl_iter)
                except StopIteration:
                    dl_iter = iter(dl)
                    X, Y = next(dl_iter)

                X = X.to(self.model.device)
                Y = Y.to(self.model.device)

                with LightwoodAutocast():
                    optimizer.zero_grad()
                    Yh = self._net_call(X)
                    loss = criterion(Yh, Y)
                    if LightwoodAutocast.active:
                        scaler.scale(loss).backward()
                        scaler.step(optimizer)
                        scaler.update()
                    else:
                        loss.backward()
                        optimizer.step()
                cum_loss += loss.item()

            log.info(f'Loss of {cum_loss} with learning rate {lr}')
            running_losses.append(cum_loss)
            lr_log.append(lr)
            cum_loss = 0
            lr = lrs.popleft()
            if len(lrs) == 0:
                stop = True

            # store model if best so far
            inv_running_losses = list(running_losses)[::-1]  # invert so when tied we pick the most aggresive LR
            best_loss_idx = np.nanargmin(inv_running_losses)  # nanargmin ignores nans that may arise
            if best_loss_idx == 0:
                best_model = deepcopy(self.model)  # store model for slight time savings
                best_loss_lr = lr_log[-1]

        lr = best_loss_lr
        log.info(f'Found learning rate of: {lr}')
        return lr, best_model

    def _max_fit(self, train_dl, dev_dl, criterion, optimizer, scaler, stop_after, return_model_after):
        epochs_to_best = 0
        best_dev_error = pow(2, 32)
        running_errors = deque(maxlen=self.loss_hist_len)
        best_model = self.model

        for epoch in range(1, return_model_after + 1):
            self.model = self.model.train()
            running_losses: List[float] = []
            for i, (X, Y) in enumerate(train_dl):
                X = X.to(self.model.device)
                Y = Y.to(self.model.device)
                with LightwoodAutocast():
                    optimizer.zero_grad()
                    Yh = self._net_call(X)
                    loss = criterion(Yh, Y)
                    if LightwoodAutocast.active:
                        scaler.scale(loss).backward()
                        scaler.step(optimizer)
                        scaler.update()
                    else:
                        loss.backward()
                        optimizer.step()

                running_losses.append(loss.item())
                if (time.time() - self.started) > stop_after:
                    break

            train_error = np.mean(running_losses)
            epoch_error = self._error(dev_dl, criterion)
            running_errors.append(epoch_error)
            log.info(f'Loss @ epoch {epoch}: {epoch_error}')

            if np.isnan(train_error) or np.isnan(
                    running_errors[-1]) or np.isinf(train_error) or np.isinf(
                    running_errors[-1]):
                break

            if best_dev_error > running_errors[-1]:
                best_dev_error = running_errors[-1]
                best_model = deepcopy(self.model)
                epochs_to_best = epoch

            # manually set epoch limit
            if self.n_epochs is not None:
                if epoch > self.n_epochs:
                    break

            # automated early stopping
            else:
                if len(running_errors) >= self.loss_hist_len:
                    delta_mean = np.average([
                        running_errors[-i - 1] - running_errors[-i] for i in range(len(running_errors) - 1)],
                        weights=[(1 / 2)**i for i in range(len(running_errors) - 1)])
                    if delta_mean >= 0:
                        break
                elif (time.time() - self.started) > stop_after:
                    break
                elif running_errors[-1] < 0.0001 or train_error < 0.0001:
                    break

        if np.isnan(best_dev_error):
            best_dev_error = pow(2, 32)
        return best_model, epochs_to_best, best_dev_error

    def _error(self, dev_dl, criterion) -> float:
        self.model = self.model.eval()
        running_losses: List[float] = []
        with torch.no_grad():
            for X, Y in dev_dl:
                X = X.to(self.model.device)
                Y = Y.to(self.model.device)
                Yh = self._net_call(X)
                running_losses.append(criterion(Yh, Y).item())
            return np.mean(running_losses)

    def _init_net(self, ds: EncodedDs):
        self.net_class = DefaultNet if self.net_name == 'DefaultNet' else ArNet

        X, Y = ds[0]
        net_kwargs = {'input_size': len(X),
                      'output_size': len(Y),
                      'num_hidden': self.num_hidden,
                      'dropout': 0}

        if self.net_class == ArNet:
            net_kwargs['encoder_span'] = ds.encoder_spans
            net_kwargs['target_name'] = self.target

        self.model = self.net_class(**net_kwargs)

    def _net_call(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    # @TODO: Compare partial fitting fully on and fully off on the benchmarks!
    # @TODO: Writeup on the methodology for partial fitting
    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the Neural mixer on some data, making it ready to predict

        :param train_data: The network is fit/trained on this
        :param dev_data: Data used for early stopping and hyperparameter determination
        """
        # ConcatedEncodedDs
        self.started = time.time()
        self.batch_size = min(200, int(len(train_data) / 10))
        self.batch_size = max(40, self.batch_size)

        dev_dl = DataLoader(dev_data, batch_size=self.batch_size, shuffle=False)
        train_dl = DataLoader(train_data, batch_size=self.batch_size, shuffle=False)

        # Find learning rate & keep initial weights
        self._init_net(train_data)
        if not self.lr:
            self.lr, self.model = self._find_lr(train_data)

        # Keep on training
        optimizer = self._select_optimizer(self.model, lr=self.lr)
        criterion = self._select_criterion()
        scaler = GradScaler()

        # Only 0.8 of the remaining time budget is used to allow some time for the final tuning and partial fit
        self.model, epoch_to_best_model, _ = self._max_fit(
            train_dl, dev_dl, criterion, optimizer, scaler, (self.stop_after - (time.time() - self.started)) * 0.8,
            return_model_after=20000)

        self.epochs_to_best += epoch_to_best_model

        if self.fit_on_dev:
            self.partial_fit(dev_data, train_data)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        self._fit(train_data, dev_data)
        self._final_tuning(dev_data)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        """
        Augments the mixer's fit with new data, nr of epochs is based on the amount of epochs the original fitting took

        :param train_data: The network is fit/trained on this
        :param dev_data: Data used for early stopping and hyperparameter determination
        """

        # Based this on how long the initial training loop took, at a low learning rate as to not mock anything up tooo badly # noqa
        self.started = time.time()
        train_dl = DataLoader(train_data, batch_size=self.batch_size, shuffle=True)
        dev_dl = DataLoader(dev_data, batch_size=self.batch_size, shuffle=True)
        optimizer = self._select_optimizer(self.model, lr=self.lr)
        criterion = self._select_criterion()
        scaler = GradScaler()

        self.model, _, _ = self._max_fit(train_dl, dev_dl, criterion, optimizer, scaler,
                                         self.stop_after * 0.1, max(1, int(self.epochs_to_best / 3)))

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Make predictions based on datasource with the same features as the ones used for fitting

        :param ds: Predictions are generate from it
        :param arg: Any additional arguments used in predicting

        :returns: A dataframe cotaining the decoded predictions and (depending on the args) additional information such as the probabilites for each target class
        """ # noqa
        self.model = self.model.eval()
        decoded_predictions: List[object] = []
        all_probs: List[List[float]] = []
        rev_map = {}

        with torch.no_grad():
            for idx, (X, Y) in enumerate(ds):
                X = X.to(self.model.device)
                Yh = self._net_call(X)
                Yh = torch.unsqueeze(Yh, 0) if len(Yh.shape) < 2 else Yh

                kwargs = {}
                for dep in self.target_encoder.dependencies:
                    kwargs['dependency_data'] = {dep: ds.data_frame.iloc[idx][[dep]].values}

                if args.predict_proba and self.supports_proba:
                    decoded_prediction, probs, rev_map = self.target_encoder.decode_probabilities(Yh, **kwargs)
                    all_probs.append(probs)
                else:
                    decoded_prediction = self.target_encoder.decode(Yh, **kwargs)

                decoded_predictions.extend(decoded_prediction)

            ydf = pd.DataFrame({'prediction': decoded_predictions})

            if args.predict_proba and self.supports_proba:
                raw_predictions = np.array(all_probs).squeeze(axis=1)

                for idx, label in enumerate(rev_map.values()):
                    ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

            return ydf



from typing import Union

import torch
import numpy as np
import pandas as pd

from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log


class DatetimeEncoder(BaseEncoder):
    """
    This encoder produces an encoded representation for timestamps.

    The approach consists on decomposing the timestamp objects into its constituent units (e.g. month, year, etc), and describing each of those with a single value that represents the magnitude in a sensible cycle length.
    """  # noqa
    def __init__(self, is_target: bool = False):
        super().__init__(is_target)
        self.constant_keys = ['year', 'month', 'day', 'hour', 'minute', 'second']
        self.constant_vals = torch.Tensor([3000.0, 12.0, 31.0, 24.0, 60.0, 60.0])  # cycle length
        self.constant_map = {k: v.item() for k, v in zip(self.constant_keys, self.constant_vals)}

        self.output_size = len(self.constant_keys)
        self.empty_vector = np.zeros((self.output_size, ))

        self.max_vals = torch.Tensor([pd.Timestamp.max.year - 1, 12, 31, 23, 59, 59])
        self.min_vals = torch.Tensor([pd.Timestamp.min.year + 1, 1, 1, 0, 0, 0])

    def prepare(self, priming_data):
        self.is_prepared = True

    def encode(self, data: Union[np.ndarray, pd.Series]) -> torch.Tensor:
        """
        :param data: a pandas series with numerical dtype, previously cleaned with dataprep_ml
        :return: encoded data, shape (len(data), self.output_size)
        """
        if type(data) not in (np.ndarray, pd.Series):
            raise Exception(f'Data should be pd.Series or np.ndarray! Got: {type(data)}')
        if isinstance(data, np.ndarray):
            data = pd.Series(data)

        data = data.fillna(pd.Timestamp.max.timestamp())  # TODO: replace with mean?
        ret = [pd.to_datetime(data, unit='s', origin=-1, utc=True)]
        for i, attr in enumerate(self.constant_keys):
            def _get_ts_attr(ts):
                return getattr(ts, attr)
            component = ret[0].apply(_get_ts_attr)
            component = component / self.constant_vals[i].item()
            ret.append(component)

        out = torch.Tensor(ret[1:])  # drop column with timestamp objects
        out = torch.transpose(out, 0, 1)  # swap dimensions to shape as (B, self.output_size)
        return out

    def decode(self, encoded_data: torch.Tensor, return_as_datetime=False) -> list:
        if len(encoded_data.shape) > 2 and encoded_data.shape[0] == 1:
            encoded_data = encoded_data.squeeze(0)

        rounded = torch.round(torch.multiply(encoded_data, self.constant_vals))
        high_bounded = torch.minimum(rounded, self.max_vals)
        low_bounded = torch.maximum(high_bounded, self.min_vals)
        ret = low_bounded.long()

        df = pd.DataFrame(ret, columns=self.constant_keys)
        nan_mask = df[
            (df['year'] == int(self.max_vals[0])) &
            (df['month'] == pd.Timestamp.max.month) &
            (df['day'] == pd.Timestamp.max.day)
        ].index
        dt = pd.to_datetime(df, utc=True)

        if not hasattr(dt, 'dt'):
            log.warning('DatetimeEncoder has failed to decode using microsecond precision, reverting to nanosecond. This may lead to minor discrepancies in reconstruction.')  # noqa

        if return_as_datetime is True:
            dt = dt.dt.to_pydatetime()  # return to Python datetime microsecond precision
            decoded = dt
        else:
            decoded = dt.values.astype(np.float64) // 10 ** 9

        decoded[nan_mask] = np.nan
        return decoded.tolist()



import librosa
import torch
import warnings
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.io import read_from_path_or_url
from lightwood.helpers.log import log
from typing import Iterable


class MFCCEncoder(BaseEncoder):
    is_trainable_encoder: bool = False

    def __init__(self, is_target: bool = False):
        """
    
        Uses `librosa` to compute the Mel-frequency spectral coefficients (MFCCs) of data within an audio file. They are a common feature used in speech and audio processing. Example: https://centaur.reading.ac.uk/88046/3/ESR_for_home_AI.pdf

        Input data to this mixer MUST BE the location of the audio files.
    
        The output feature for any given audio file is a 2D array, flattened into a 1D one to comply with the expected format in lightwood mixers.
    
        This encoder currently does not support a `decode()` call; models with an audio output will not work. 
    
        :param is_target: whether this encoder's column is the target. Is always false, as decoded audio is not available.
        """  # noqa
        assert not is_target
        super().__init__(is_target)

    def prepare(self, priming_data: Iterable[str]):
        """
        Audio-encoding is rule-based; prepare only sets the size of the output dimension based on encoding the first example within the priming data.

        :param priming_data: Training data, in the form of file locations
        """  # noqa
        self.is_prepared = True
        priming_data = list(priming_data)
        ele = self.encode([str(priming_data[0])])[0]
        self.output_size = len(ele)

    def encode(self, column_data: Iterable[str]) -> torch.Tensor:
        """
        Encode a list of audio files via mfccs.

        :param column_data: list of strings that point to paths or URLs of the audio files that will be encoded.
        """
        encoded_audio_arr = []
        for path in column_data:
            try:
                y, _ = read_from_path_or_url(path, librosa.load)
            except Exception as e:
                log.error(f'Unable to read audio file {path}, error: {e}')
                encoded_audio_arr.append([0] * self.output_size)
                continue

            # If the durations of the audio samples are highly variable, the
            # same coefficients will refer to time buckets of different lenghts.
            # This means that a model will find difficult to use temporal
            # information
            NUM_TIME_BUCKETS = 100  # audio file will be split into 100 sequential time intervals before computing the Fourier transform needed for the MFCCs. A value of `100` will split a 1s audio file in 10ms intervals, which are enough for speech recognition. It will also split a 3 minutes song in 1.8s intervals, which are still small enough to capture enough detail for genre recognition  # noqa
            N_MFCC_COEFFICIENTS = 20

            num_samples = y.shape[0]

            # truncate towards 1
            hop_length = int(num_samples / NUM_TIME_BUCKETS + 1)

            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=DeprecationWarning)
                mfcc_coefficients = librosa.feature.mfcc(
                    y,
                    n_mfcc=N_MFCC_COEFFICIENTS,
                    hop_length=hop_length
                ).reshape(-1)

            encoded_audio_arr.append(mfcc_coefficients)

        return torch.Tensor(encoded_audio_arr)

    def decode(self, _):
        raise Exception('This encoder is not bi-directional')



from typing import List, Optional

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from type_infer.helpers import is_nan_numeric
from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from mindsdb_evaluator import evaluate_accuracies
from type_infer.dtype import dtype


class WeightedMeanEnsemble(BaseEnsemble):
    """
    This ensemble determines a weight vector to return a weighted mean of the underlying mixers.

    More specifically, each model is evaluated on the validation dataset and assigned an accuracy score (as per the fixed accuracy function at the JsonAI level).

    Afterwards, all the scores are softmaxed to obtain the final weights.

    Note: this ensemble only supports regression tasks.
    """  # noqa
    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, args: PredictionArguments,
                 dtype_dict: dict, accuracy_functions, ts_analysis: Optional[dict] = None,
                 fit: Optional[bool] = True, **kwargs) -> None:
        super().__init__(target, mixers, data)
        if dtype_dict[target] not in (dtype.float, dtype.integer, dtype.quantity):
            raise Exception(
                f'This ensemble can only be used regression problems! Got target dtype {dtype_dict[target]} instead!')

        if fit:
            score_list = []
            for _, mixer in enumerate(mixers):
                score_dict = evaluate_accuracies(
                    data.data_frame,
                    mixer(data, args)['prediction'],
                    target,
                    accuracy_functions,
                    ts_analysis=ts_analysis
                )
                avg_score = np.mean(list(score_dict.values()))
                log.info(f'Mixer: {type(mixer).__name__} got accuracy: {avg_score}')

                if is_nan_numeric(avg_score):
                    log.warning(f'Could not compute a valid accuracy for mixer: {type(mixer).__name__}, \
                                  functions: {accuracy_functions}, yielded invalid average score {avg_score}, \
                                  resetting that to -pow(2,63) instead.')
                    avg_score = -pow(2, 63)

                score_list.append(avg_score)

            self.weights = self.accuracies_to_weights(np.array(score_list))
            self.prepared = True

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        assert self.prepared
        df = pd.DataFrame()
        for mixer in self.mixers:
            df[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']

        mixer_weights = args.mixer_weights if args.mixer_weights else self.weights
        avg_predictions_df = df.apply(lambda x: np.average(x, weights=mixer_weights), axis='columns')
        return pd.DataFrame(avg_predictions_df, columns=['prediction'])

    @staticmethod
    def accuracies_to_weights(x: np.array) -> np.array:
        # Converts accuracies to weights using the softmax function.
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum()



from typing import List, Dict, Iterable, Optional

import torch

from lightwood.encoder import BaseEncoder
from lightwood.encoder.numeric import TsNumericEncoder


class TsArrayNumericEncoder(BaseEncoder):
    def __init__(self, timesteps: int, is_target: bool = False, positive_domain: bool = False, grouped_by=None, nan=0):
        """
        This encoder handles arrays of numerical time series data by wrapping the numerical encoder with behavior specific to time series tasks.
        
        :param timesteps: length of forecasting horizon, as defined by TimeseriesSettings.window.
        :param is_target: whether this encoder corresponds to the target column.
        :param positive_domain: whether the column domain is expected to be positive numbers.
        :param grouped_by: what columns, if any, are considered to group the original column and yield multiple time series.
        """  # noqa
        super(TsArrayNumericEncoder, self).__init__(is_target=is_target)
        self.normalizers = None
        self.group_combinations = None
        self.dependencies = grouped_by
        self.data_window = timesteps
        self.positive_domain = positive_domain
        self.nan_value = nan
        self.sub_encoder = TsNumericEncoder(is_target=is_target, positive_domain=positive_domain, grouped_by=grouped_by)
        self.output_size = self.data_window * self.sub_encoder.output_size

    def prepare(self, priming_data):
        """
        This method prepares the underlying time series numerical encoder.
        """
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self.sub_encoder.prepare(priming_data)
        self.is_prepared = True

    def encode(self, data: Iterable[Iterable], dependency_data: Optional[Dict[str, str]] = {}) -> torch.Tensor:
        """
        Encodes a list of time series arrays using the underlying time series numerical encoder.
        
        :param data: list of numerical values to encode. Its length is determined by the tss.window parameter, and all data points belong to the same time series.
        :param dependency_data: dict with values of each group_by column for the time series, used to retrieve the correct normalizer.
        
        :return: list of encoded time series arrays. Tensor is (len(data), N x K)-shaped, where N: self.data_window and K: sub-encoder # of output features.
        """  # noqa
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')
        if self.sub_encoder.normalizers is None and self.normalizers is not None:
            self.sub_encoder.normalizers = self.normalizers
        if not dependency_data:
            dependency_data = {'__default': [None] * len(data)}

        ret = self.sub_encoder.encode(data, dependency_data=dependency_data)

        return torch.Tensor(ret).nan_to_num(self.nan_value)

    def decode(self, encoded_values, dependency_data=None) -> List[List]:
        """
        Decodes a list of encoded arrays into values in their original domains.

        :param encoded_values: encoded slices of numerical time series.
        :param dependency_data: used to determine the correct normalizer for the input.

        :return: a list of decoded time series arrays.
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        encoded_values = encoded_values.reshape(encoded_values.shape[0],
                                                self.data_window,
                                                self.sub_encoder.output_size)

        ret = []
        for tensor in torch.split(encoded_values, 1, dim=0):
            ret.append(self.decode_one(tensor, dependency_data=dependency_data))

        return ret

    def decode_one(self, encoded_value, dependency_data={}) -> List:
        """
        Decodes a single window of a time series into its original domain.

        :param encoded_value: encoded slice of a numerical time series.
        :param dependency_data: used to determine the correct normalizer for the input.

        :return: a list of length TimeseriesSettings.window with decoded values for the forecasted time series.
        """
        ret = []
        for encoded_timestep in torch.split(encoded_value, 1, dim=1):
            ret.extend(self.sub_encoder.decode(encoded_timestep.squeeze(1), dependency_data=dependency_data))
        return ret



from typing import List, Optional

import numpy as np
import pandas as pd

import torch
from torch import nn
from torch.optim import SGD

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.mean_ensemble import MeanEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from lightwood.helpers.log import log


class StackedEnsemble(MeanEnsemble):
    """
    This ensemble will learn an optimal weight vector via Stochastic Gradient Descent on the validation dataset and the respective mixer predictions.

    Starting weights for the vector are uniformly set.

    Note this mixer is still in experimental phase. Some features in the roadmap are:
      - support for handling faulty mixers
      - support for custom initial vector weights
      - early stopping
      - arbitrarily complex secondary model

    """  # noqa
    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, dtype_dict: dict,
                 args: PredictionArguments, fit: Optional[bool] = True, **kwargs) -> None:
        super().__init__(target, mixers, data, dtype_dict, fit=False)

        self.target_cols = [target]
        self.softmax = nn.Softmax(dim=0)
        self.mixer_weights = nn.Parameter(torch.full((len(mixers),), 1 / len(mixers)))
        self.criterion = nn.MSELoss()
        self.optimizer = SGD([self.mixer_weights], lr=0.01)
        self.agg_dim = 1

        if fit and len(mixers) > 1:
            all_preds = torch.tensor(self.predict(data, args)).squeeze().reshape(-1, len(mixers))
            actual = torch.tensor(data.data_frame[self.target_cols].values)

            def _eval_loss():
                self.optimizer.zero_grad()
                weighted = torch.sum(all_preds * self.softmax(self.mixer_weights), dim=self.agg_dim)
                loss = self.criterion(weighted, actual)
                loss.backward()
                return loss

            self.optimizer.step(_eval_loss)
            self.mixer_weights = self.softmax(self.mixer_weights)
            log.info(f'Optimal stacking weights: {self.mixer_weights.detach().tolist()}')
        self.prepared = True

    def predict(self, ds: EncodedDs, args: PredictionArguments) -> List:
        outputs = []
        for mixer in self.mixers:
            output = mixer(ds, args=args)['prediction'].tolist()
            output = np.expand_dims(np.array(output), self.agg_dim)
            outputs.append(output)
        return outputs

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        assert self.prepared
        mixer_weights = torch.tensor(args.mixer_weights) if args.mixer_weights else self.mixer_weights
        output = pd.DataFrame()
        predictions = torch.tensor(self.predict(ds, args)).squeeze().reshape(-1, len(self.mixers))
        predictions = (predictions * mixer_weights).sum(axis=1)
        output['prediction'] = predictions.detach().numpy().tolist()
        return output

    def set_weights(self, weights: List):
        if len(weights) != len(self.mixers):
            raise Exception(f"Expected weight vector to have {len(self.mixers)} entries, got {len(weights)} instead.")

        self.mixer_weights = torch.tensor(weights)



from typing import List, Iterable

import torch

from type_infer.dtype import dtype
from lightwood.encoder.array import ArrayEncoder


class TimeSeriesEncoder(ArrayEncoder):
    is_timeseries_encoder: bool = True
    is_trainable_encoder: bool = True

    def __init__(self, stop_after: float, window: int = None, is_target: bool = False, original_type: dtype = None):
        """
        Time series encoder. This module will pass the normalized series values, along with moving averages taken from the series' last `window` values.
        :param stop_after: time budget in seconds.
        :param window: expected length of array data.
        :param original_type: element-wise data type
        """  # noqa
        super().__init__(stop_after, window, is_target, original_type)
        self.max_mavg_offset = self.output_size
        self.output_size += self.max_mavg_offset

    def encode(self, column_data: Iterable[Iterable]) -> torch.Tensor:
        """
        Encodes time series data.

        :param column_data: Input column data to be encoded
        :returns: a torch tensor representing the encoded time series.
        """

        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        base_encode = super().encode(column_data)

        if self.original_type in (dtype.integer, dtype.float, dtype.quantity, dtype.num_tsarray):
            mavgs = []
            for offset in range(self.max_mavg_offset):
                ma = torch.mean(base_encode[:, offset:self.max_mavg_offset], 1)
                mavgs.append(ma.unsqueeze(1))
            base_encode[:, (self.output_size - self.max_mavg_offset):] = torch.cat(mavgs, dim=1)

        return base_encode

    def decode(self, data: torch.Tensor) -> List[Iterable]:
        """
        Converts data as a list of arrays. Removes all encoded moving average information.

        :param data: Encoded data prepared by this array encoder
        :returns: A list of iterable sequences in the original data space
        """
        decoded = data[:, self.max_mavg_offset].tolist()
        return decoded



from typing import Dict, List, Tuple, Optional

import numpy as np
from dataprep_ml import StatisticalAnalysis

from lightwood.helpers.log import log
from type_infer.dtype import dtype
from lightwood.ensemble import BaseEnsemble
from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.data.encoded_ds import EncodedDs
from lightwood.encoder.text.pretrained import PretrainedLangEncoder
from lightwood.api.types import ModelAnalysis, ProblemDefinition, PredictionArguments


def model_analyzer(
    predictor: BaseEnsemble,
    data: EncodedDs,
    train_data: EncodedDs,
    stats_info: StatisticalAnalysis,
    target: str,
    pdef: ProblemDefinition,
    dtype_dict: Dict[str, str],
    accuracy_functions,
    ts_analysis: Dict,
    analysis_blocks: Optional[List[BaseAnalysisBlock]] = []
) -> Tuple[ModelAnalysis, Dict[str, object]]:
    """
    Analyses model on a validation subset to evaluate accuracy, estimate feature importance and generate a
    calibration model to estimating confidence in future predictions.

    Additionally, any user-specified analysis blocks (see class `BaseAnalysisBlock`) are also called here.

    :return:
    runtime_analyzer: This dictionary object gets populated in a sequential fashion with data generated from
    any `.analyze()` block call. This dictionary object is stored in the predictor itself, and used when
    calling the `.explain()` method of all analysis blocks when generating predictions.

    model_analysis: `ModelAnalysis` object that contains core analysis metrics, not necessarily needed when predicting.
    """

    runtime_analyzer = {}
    data_type = dtype_dict[target]
    tss = pdef.timeseries_settings

    # retrieve encoded data representations
    encoded_train_data = train_data
    encoded_val_data = data
    data = encoded_val_data.data_frame
    input_cols = list([col for col in data.columns if col != target])

    if not pdef.embedding_only:
        # predictive task
        is_numerical = data_type in (dtype.integer, dtype.float, dtype.num_tsarray, dtype.quantity)
        is_classification = data_type in (dtype.categorical, dtype.binary, dtype.cat_tsarray)
        is_multi_ts = tss.is_timeseries and tss.horizon > 1
        has_pretrained_text_enc = any([isinstance(enc, PretrainedLangEncoder)
                                       for enc in encoded_train_data.encoders.values()])

        # raw predictions for validation dataset
        args = {} if not is_classification else {"predict_proba": True}
        normal_predictions = None

        if len(analysis_blocks) > 0:
            if tss.is_timeseries:
                # we retrieve the first entry per group (closest to supervision cutoff)
                if tss.group_by:
                    encoded_val_data.data_frame['__mdb_val_idx'] = np.arange(len(encoded_val_data))
                    idxs = encoded_val_data.data_frame.groupby(by=tss.group_by).first()['__mdb_val_idx'].values
                    encoded_val_data.data_frame = encoded_val_data.data_frame.iloc[idxs, :]
                    if encoded_val_data.cache_built:
                        encoded_val_data.X_cache = encoded_val_data.X_cache[idxs, :]
                        encoded_val_data.Y_cache = encoded_val_data.Y_cache[idxs, :]
            normal_predictions = predictor(encoded_val_data, args=PredictionArguments.from_dict(args))
            normal_predictions = normal_predictions.set_index(encoded_val_data.data_frame.index)

        # ------------------------- #
        # Run analysis blocks, both core and user-defined
        # ------------------------- #
        kwargs = {
            'predictor': predictor,
            'target': target,
            'input_cols': input_cols,
            'dtype_dict': dtype_dict,
            'normal_predictions': normal_predictions,
            'data': encoded_val_data.data_frame,
            'train_data': train_data,
            'encoded_val_data': encoded_val_data,
            'is_classification': is_classification,
            'is_numerical': is_numerical,
            'is_multi_ts': is_multi_ts,
            'stats_info': stats_info,
            'tss': tss,
            'ts_analysis': ts_analysis,
            'accuracy_functions': accuracy_functions,
            'has_pretrained_text_enc': has_pretrained_text_enc
        }

        for block in analysis_blocks:
            log.info("The block %s is now running its analyze() method", block.__class__.__name__)
            runtime_analyzer = block.analyze(runtime_analyzer, **kwargs)

    # ------------------------- #
    # Populate ModelAnalysis object
    # ------------------------- #
    model_analysis = ModelAnalysis(
        accuracies=runtime_analyzer.get('score_dict', {}),
        accuracy_histogram=runtime_analyzer.get('acc_histogram', {}),
        accuracy_samples=runtime_analyzer.get('acc_samples', {}),
        train_sample_size=len(encoded_train_data),
        test_sample_size=len(encoded_val_data),
        confusion_matrix=runtime_analyzer.get('cm', []),
        column_importances=runtime_analyzer.get('column_importances', {}),
        histograms=stats_info.histograms,
        dtypes=dtype_dict,
        submodel_data=predictor.submodel_data
    )

    return model_analysis, runtime_analyzer



from typing import Dict, Optional

from lightwood.mixer.sktime import SkTime


class ARIMAMixer(SkTime):
    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            horizon: int,
            ts_analysis: Dict,
            model_path: str = 'statsforecast.StatsForecastAutoARIMA',
            auto_size: bool = True,
            sp: int = None,
            use_stl: bool = False,
            start_p: int = 2,
            d: Optional[int] = None,
            start_q: int = 2,
            max_p: int = 5,
            max_d: int = 2,
            max_q: int = 5,
            start_P: int = 1,
            D: Optional[int] = None,
            start_Q: int = 1,
            max_P: int = 2,
            max_D: int = 1,
            max_Q: int = 2,
            max_order: int = 5,
            seasonal: bool = True,
            stationary: bool = False,
            information_criterion: Optional[str] = None,
            alpha: float = 0.05,
            test: str = 'kpss',
            seasonal_test: Optional[str] = None,
            stepwise: bool = True,
            n_jobs: Optional[int] = None,
            start_params: Optional[list] = None,
            trend: Optional[str] = None,
            method: Optional[str] = None,
            maxiter: int = 50,
            offset_test_args: Optional[dict] = None,
            seasonal_test_args: Optional[dict] = None,
            suppress_warnings: bool = True,
            error_action: str = 'warn',
            trace: bool = False,
            random: bool = False,
            random_state: Optional[int] = None,
            n_fits: Optional[int] = None,
            out_of_sample_size: int = 0,
            scoring: str = 'mse',
            scoring_args: Optional[dict] = None,
            with_intercept: bool = True,
            update_pdq: bool = True,
            time_varying_regression: bool = False,
            enforce_stationarity: bool = True,
            enforce_invertibility: bool = True,
            simple_differencing: bool = False,
            measurement_error: bool = False,
            mle_regression: bool = True,
            hamilton_representation: bool = False,
            concentrate_scale: bool = False,
    ):
        """
        Wrapper for SkTime's AutoARIMA interface.

        :param stop_after: time budget in seconds
        :param target: column containing target time series
        :param dtype_dict: data types for each dataset column
        :param horizon: forecast length
        :param ts_analysis: lightwood-produced stats about input time series
        :param auto_size: whether to filter out old data points if training split is bigger than a certain threshold (defined by the dataset sampling frequency). Enabled by default to avoid long training times in big datasets.
        :param use_stl: Whether to use de-trenders and de-seasonalizers fitted in the timeseries analysis phase.

        For the rest of the parameters, please refer to SkTime's documentation.
        """  # noqa
        hyperparam_search = False
        model_kwargs = {
            'start_p': start_p,
            'd': d,
            'start_q': start_q,
            'max_p': max_p,
            'max_d': max_d,
            'max_q': max_q,
            'start_P': start_P,
            'D': D,
            'start_Q': start_Q,
            'max_P': max_P,
            'max_D': max_D,
            'max_Q': max_Q,
            'max_order': max_order,
            'seasonal': seasonal,
            'stationary': stationary,
            'information_criterion': information_criterion,
            'alpha': alpha,
            'test': test,
            'seasonal_test': seasonal_test,
            'stepwise': stepwise,
            'n_jobs': n_jobs,
            'start_params': start_params,
            'trend': trend,
            'method': method,
            'maxiter': maxiter,
            'offset_test_args': offset_test_args,
            'seasonal_test_args': seasonal_test_args,
            'suppress_warnings': suppress_warnings,
            'error_action': error_action,
            'trace': trace,
            'random': random,
            'random_state': random_state,
            'n_fits': n_fits,
            'out_of_sample_size': out_of_sample_size,
            'scoring': scoring,
            'scoring_args': scoring_args,
            'with_intercept': with_intercept,
            'update_pdq': update_pdq,
            'time_varying_regression': time_varying_regression,
            'enforce_stationarity': enforce_stationarity,
            'enforce_invertibility': enforce_invertibility,
            'simple_differencing': simple_differencing,
            'measurement_error': measurement_error,
            'mle_regression': mle_regression,
            'hamilton_representation': hamilton_representation,
            'concentrate_scale': concentrate_scale,
        }
        super().__init__(stop_after, target, dtype_dict, horizon, ts_analysis,
                         model_path, model_kwargs, auto_size, sp, hyperparam_search, use_stl)
        self.name = 'AutoARIMA'
        self.stable = False



from lightwood.encoder.datetime.datetime import DatetimeEncoder
from lightwood.encoder.datetime.datetime_sin_normalizer import DatetimeNormalizerEncoder

__all__ = ['DatetimeEncoder', 'DatetimeNormalizerEncoder']



from typing import Dict, Optional

from lightwood.mixer.sktime import SkTime


class ETSMixer(SkTime):
    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            horizon: int,
            ts_analysis: Dict,
            model_path: str = 'ets.AutoETS',
            auto_size: bool = True,
            sp: int = None,
            use_stl: bool = True,
            error: str = 'add',
            trend: Optional[str] = None,
            damped_trend: bool = False,
            seasonal: Optional[str] = None,
            initialization_method: str = 'estimated',
            initial_level: Optional[float] = None,
            initial_trend: Optional[float] = None,
            initial_seasonal: Optional[list] = None,
            bounds: Optional[dict] = None,
            start_params: Optional[list] = None,
            maxiter: int = 1000,
            auto: bool = False,
            information_criterion: str = 'aic',
            allow_multiplicative_trend: bool = False,
            restrict: bool = True,
            additive_only: bool = False,
            ignore_inf_ic: bool = True,
            n_jobs: Optional[int] = None,
            random_state: Optional[int] = None
    ):
        """
        Wrapper for SkTime's AutoETS interface.

        :param stop_after: time budget in seconds
        :param target: column containing target time series
        :param dtype_dict: data types for each dataset column
        :param horizon: forecast length
        :param ts_analysis: lightwood-produced stats about input time series
        :param auto_size: whether to filter out old data points if training split is bigger than a certain threshold (defined by the dataset sampling frequency). Enabled by default to avoid long training times in big datasets.
        :param use_stl: Whether to use de-trenders and de-seasonalizers fitted in the timeseries analysis phase.

        For the rest of the parameters, please refer to SkTime's documentation.
        """  # noqa

        hyperparam_search = False
        model_kwargs = {
            'error': error,
            'trend': trend,
            'damped_trend': damped_trend,
            'seasonal': seasonal,
            'initialization_method': initialization_method,
            'initial_level': initial_level,
            'initial_trend': initial_trend,
            'initial_seasonal': initial_seasonal,
            'bounds': bounds,
            'start_params': start_params,
            'maxiter': maxiter,
            'auto': auto,
            'information_criterion': information_criterion,
            'allow_multiplicative_trend': allow_multiplicative_trend,
            'restrict': restrict,
            'additive_only': additive_only,
            'ignore_inf_ic': ignore_inf_ic,
            'n_jobs': n_jobs,
            'random_state': random_state
        }
        super().__init__(stop_after, target, dtype_dict, horizon, ts_analysis,
                         model_path, model_kwargs, auto_size, sp, hyperparam_search, use_stl)
        self.name = 'AutoETS'
        self.stable = False



import torch
import qiskit
import numpy as np

from lightwood.mixer.helpers.default_net import DefaultNet
from lightwood.helpers.torch import LightwoodAutocast


class QuantumCircuit:
    """
    This class provides a simple interface for interaction
    with the quantum circuit
    """

    def __init__(self, n_qubits, backend, shots):
        # --- Circuit definition ---
        self._circuit = qiskit.QuantumCircuit(n_qubits)

        all_qubits = [i for i in range(n_qubits)]
        self.theta = qiskit.circuit.Parameter('theta')

        self._circuit.h(all_qubits)
        self._circuit.barrier()
        self._circuit.ry(self.theta, all_qubits)

        self._circuit.measure_all()
        # ---------------------------

        self.backend = backend
        self.shots = shots

    def run(self, thetas):
        job = self.backend.run([self._circuit.bind_parameters({self.theta: t})
                                for t in thetas], shots=self.shots)
        results = job.result().get_counts()

        final = []
        for result in results:
            counts = np.array(list(result.values()))
            states = np.array(list(result.keys())).astype(float)

            # Compute probabilities for each state
            probabilities = counts / self.shots

            # Get state expectation
            expectation = np.sum(states * probabilities)
            final.append(expectation)
        return np.array(final)


class HybridSingleFunction(torch.autograd.Function):
    """ Hybrid quantum - classical function definition """

    @staticmethod
    def forward(ctx, input, quantum_circuit, shift):
        """ Forward pass computation """
        ctx.shift = shift
        ctx.quantum_circuit = quantum_circuit

        expectation_z = ctx.quantum_circuit.run(input.tolist())
        result = torch.tensor(expectation_z)
        ctx.save_for_backward(input, result)

        return result

    @staticmethod
    def backward(ctx, grad_output):
        """ Backward pass computation """
        input, expectation_z = ctx.saved_tensors
        input_list = np.array(input.tolist())

        shift_right = input_list + np.ones(input_list.shape) * ctx.shift
        shift_left = input_list - np.ones(input_list.shape) * ctx.shift

        expectation_left = ctx.quantum_circuit.run(shift_left)
        expectation_right = ctx.quantum_circuit.run(shift_right)
        gradients = torch.tensor([expectation_right]) - torch.tensor([expectation_left])
        gradients = np.array(gradients).T
        return torch.tensor(gradients).float() * grad_output.float(), None, None


class HybridSingle(torch.nn.Module):
    """ Hybrid quantum - classical layer definition """

    def __init__(self, backend, shots, shift):
        super(HybridSingle, self).__init__()
        self.quantum_circuit = QuantumCircuit(1, backend, shots)
        self.shift = shift

    def forward(self, input):
        return HybridSingleFunction.apply(input, self.quantum_circuit, self.shift)


class QClassicNet(DefaultNet):
    """
    DefaultNet variant that uses qiskit to add a final quantum layer
    """

    def __init__(self,
                 input_size: int = None,
                 output_size: int = None,
                 shape: list = None,
                 max_params: int = 3e7,
                 num_hidden: int = 1,
                 dropout: float = 0) -> None:
        hidden_size = max([input_size * 2, output_size * 2, 400])
        super().__init__(input_size=input_size,
                         output_size=hidden_size,
                         shape=shape,
                         max_params=max_params,
                         num_hidden=num_hidden,
                         dropout=dropout)
        self.fc = torch.nn.Linear(hidden_size, output_size)
        self.hybrid = HybridSingle(qiskit.Aer.get_backend('aer_simulator'), 100, np.pi / 2)

    def to(self, device=None, available_devices=None):
        return super().to(device)

    def forward(self, input):
        with LightwoodAutocast():
            if len(input.shape) == 1:
                input = input.unsqueeze(0)
            classical_output = self.fc(self.net(input))
            full_output = torch.stack([self.hybrid(i) for i in classical_output])
        return full_output.float()



from typing import List, Optional, Dict

import pandas as pd
import numpy as np

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from type_infer.dtype import dtype
from mindsdb_evaluator import evaluate_accuracies
from type_infer.helpers import is_nan_numeric
from lightwood.helpers.log import log


class ModeEnsemble(BaseEnsemble):
    """
    When called, this ensemble will return the mode prediction from the entire list of underlying mixers.

    If there are multiple modes, the mode whose voting mixers have the highest score will be returned.

    NOTE: can only be used in categorical tasks.
    """
    mixer_scores: Dict[str, float]

    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, dtype_dict: dict,
                 accuracy_functions, args: PredictionArguments, ts_analysis: Optional[dict] = None,
                 fit: Optional[bool] = True, **kwargs) -> None:
        super().__init__(target, mixers, data, fit=False)
        self.mixer_scores = {}

        if fit:
            if dtype_dict[target] not in (dtype.binary, dtype.categorical, dtype.tags):
                raise Exception(
                    'This ensemble can only be used in classification problems! ' +
                    f'Got target dtype {dtype_dict[target]} instead!')

            for _, mixer in enumerate(mixers):
                score_dict = evaluate_accuracies(
                    data.data_frame,
                    mixer(data, args)['prediction'],
                    target,
                    accuracy_functions,
                    ts_analysis=ts_analysis
                )
                avg_score = np.mean(list(score_dict.values()))
                log.info(f'Mixer: {type(mixer).__name__} got accuracy: {avg_score}')

                if is_nan_numeric(avg_score):
                    avg_score = -pow(2, 63)
                    log.warning(f'Change the accuracy of mixer {type(mixer).__name__} to valid value: {avg_score}')

                self.mixer_scores[f'__mdb_mixer_{type(mixer).__name__}'] = avg_score
            self.prepared = True

    def _pick_mode_highest_score(self, prediction: pd.Series):
        """If the predictions are unimodal, return the mode. If there are multiple modes, return the mode whose voting
        mixers have the highest score."""
        prediction_counts = prediction.value_counts()

        # If there is a clear winner, i.e. only one prediction
        if len(prediction_counts) == 1:
            return prediction_counts.index[0]

        counts = prediction_counts.values  # how many times all predictions have appeared
        max_count = np.max(counts)  # how many times the most frequent predictions have apppeared

        # most frequent predictions and how many times they appeared
        modes = prediction_counts[prediction_counts == max_count]

        modes_predictions = modes.index  # most frequent predictions

        # For each mode, get the sum of the scores of the predictors who voted for it
        modes_predictions_scores = {}
        for mode_prediction in modes_predictions:
            voting_mixers_name = prediction[prediction == mode_prediction].index.tolist()
            modes_predictions_scores[mode_prediction] = np.sum(
                [self.mixer_scores[mixer_name] for mixer_name in voting_mixers_name])

        # Return the mode with the maximum sum of accuracies
        return max(modes_predictions_scores, key=modes_predictions_scores.get)

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        assert self.prepared
        predictions_df = pd.DataFrame()
        for mixer in self.mixers:
            predictions_df[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']

        mode_df = predictions_df.apply(func=self._pick_mode_highest_score, axis='columns')

        return pd.DataFrame(mode_df, columns=['prediction'])



from typing import Iterable

import numpy as np
from type_infer.helpers import is_nan_numeric


def is_none(value):
    """
    Pandas has no way to guarantee "stability" for the type of a column, it choses to arbitrarily change it based on the values.
    Pandas also change the values in the columns based on the types.
    Lightwood relies on having ``None`` values for a cells that represent "missing" or "corrupt".
    
    When we assign ``None`` to a cell in a dataframe this might get turned to `nan` or other values, this function checks if a cell is ``None`` or any other values a pd.DataFrame might convert ``None`` to.

    It also checks some extra values (like ``''``) that pandas never converts ``None`` to (hopefully). But lightwood would still consider those values "None values", and this will allow for more generic use later.
    """ # noqa
    # TODO: consider removing this helper entirely, arguably should move into dataprep_ml

    # start dispatch with types that are expensive to e.g. cast into strings
    if type(value) in (np.ndarray,) and value.size == 0:
        return True

    if type(value) != str and isinstance(value, Iterable) and len(value) == 0:
        return True
    elif type(value) != str and isinstance(value, Iterable):
        return False

    if value is None:
        return True

    if is_nan_numeric(value):
        return True

    if str(value) == '':
        return True

    if str(value) in ('None', 'nan', 'NaN', 'np.nan', 'NAN', 'NONE'):
        return True

    return False




import numpy as np
from typing import List, Iterable, Union
import torch
from lightwood.encoder.base import BaseEncoder


class IdentityEncoder(BaseEncoder):
    """
    Preserves the original input data without any transformations.

    Due to the restrictions of torch.Tensor, the encoding function only accepts types:

    - (nested) lists of numbers
    - np.ndarray of numbers, or
    - torch.Tensor

    Nonetypes are automatically converted into nans. This cannot be decoded.
    If self.handle_nan is True, then all nans are converted to zeros and not decoded back.

    The decode function takes in a torch.Tensor and converts it to a list of numbers.
    """  # noqa

    def __init__(self, is_target: bool = False, handle_nan: bool = True) -> None:
        """
        :param is_target: Whether encoder represents the target column
        :param handle_nan: If True, converts any NaN values to 0 via `torch.nan_to_num`.
        """
        super().__init__(is_target)
        self.handle_nan = handle_nan

    # Not all encoders need to be prepared
    def prepare(self, priming_data: Iterable[Union[float, int]]) -> None:
        """ Nothing to prepare """
        self._prepared = True

    def encode(self, column_data: Iterable[Union[float, int]]) -> torch.Tensor:
        """
        Converts input data into a torch tensor with NaN handling, if necessary.

        :param column_data: Input iterable of numerical data
        :returns: Torch tensor of the data type
        """
        if isinstance(column_data, torch.Tensor):
            res = column_data
        else:
            res = np.array(column_data, dtype=float)  # converts None to nan
            res = torch.Tensor(res)

        if self.handle_nan:
            res = torch.nan_to_num(res)
        return res

    def decode(self, encoded_data: torch.Tensor) -> List[object]:
        """ Returns a list of the input data """
        return encoded_data.tolist()



from typing import Union

import numpy as np
import pandas as pd
from scipy.stats import entropy
from sklearn.linear_model import Ridge
from sklearn.metrics import mean_absolute_error

from type_infer.dtype import dtype
from lightwood.mixer import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class Normalizer(BaseMixer):
    """
    Companion class to the confidence estimation analysis block. A normalizer is a secondary machine learning model
    tasked with learning to estimate the "difficulty" that the main predictor will have with any problem instance.

    The idea is that this model should emit higher scores for tougher predictions. All scores will be passed as a
    normalizing factor to the conformal prediction framework, thus:
      - widening bounds at the same confidence level if a prediction is harder
      - tightening bounds at the same confidence level if a predictions is easier
      
    Reference:
        Papadopoulos, H., Gammerman, A., & Vovk, V. (2008). Normalized nonconformity measures for regression Conformal Prediction.

    """  # noqa
    def __init__(self, fit_params: dict):
        super(Normalizer, self).__init__(stop_after=fit_params['stop_after'])

        self.input_cols = list(fit_params['dtype_dict'].keys())
        self.base_predictor = fit_params['predictor']
        self.encoders = fit_params['encoders']
        self.target = fit_params['target']
        self.target_dtype = fit_params['dtype_dict'][fit_params['target']]
        self.multi_ts_task = fit_params['is_multi_ts']

        self.model = Ridge()  # TODO: enable underlying model selection from JsonAI
        self.prepared = False
        self.prediction_cache = None
        self.bounds = (0.5, 1.5)
        self.error_fn = mean_absolute_error

    def fit(self, data: EncodedDs) -> None:
        try:
            data = ConcatedEncodedDs([data])
            preds = self.base_predictor(data, args=PredictionArguments.from_dict({'predict_proba': True}))
            truths = data.data_frame[self.target]
            labels = self.get_labels(preds, truths.values, data.encoders[self.target])
            enc_data = data.get_encoded_data(include_target=False).numpy()
            self.model.fit(enc_data, labels)
            self.prepared = True
        except Exception:
            pass

    def __call__(self, ds: Union[ConcatedEncodedDs, EncodedDs, np.ndarray], args: PredictionArguments) \
            -> np.ndarray:
        if isinstance(ds, EncodedDs) or isinstance(ds, ConcatedEncodedDs):
            ds = ds.get_encoded_data(include_target=False).numpy()  # TODO: set upper limit for speed

        if self.prepared:
            raw = self.model.predict(ds)
            scores = np.clip(raw, 0.1, 1e4)  # set limit deviations (TODO: benchmark stability)
            # scores = scores / scores.mean()  # smoothed
        else:
            scores = np.ones(ds.shape[0])

        return scores

    def score(self, data) -> np.ndarray:
        if not self.prepared:
            scores = np.ones(data.shape[0])  # by default, normalizing factor is 1 for all predictions
        elif self.prediction_cache is not None:
            scores = self.prediction_cache
        else:
            scores = self.model.predict(data)

        return scores

    def get_labels(self, preds: pd.DataFrame, truths: np.ndarray, target_enc) -> np.ndarray:
        if self.target_dtype in [dtype.integer, dtype.float, dtype.quantity]:
            if not self.multi_ts_task:
                preds = preds.values.squeeze()
            else:
                preds = pd.Series([p[0] for p in preds.values.squeeze()]).values
            preds = preds.astype(float)
            labels = self.compute_numerical_labels(preds, truths, self.bounds)

        elif self.target_dtype in [dtype.binary, dtype.categorical]:
            if self.base_predictor.supports_proba:
                prob_cols = [col for col in preds.columns if '__mdb_proba' in col]
                col_names = [col.replace('__mdb_proba_', '') for col in prob_cols]
                preds = preds[prob_cols]
            else:
                prob_cols = col_names = target_enc.map.keys()
                ohe_preds = pd.get_dummies(preds['prediction'], columns=col_names)
                for col in col_names:
                    if col not in ohe_preds.columns:
                        ohe_preds[col] = np.zeros(ohe_preds.shape[0])
                preds = ohe_preds

            # reorder preds to ensure classes are in same order as in target_enc
            preds.columns = col_names
            new_order = [v for k, v in sorted(target_enc.rev_map.items(), key=lambda x: x[0])]
            preds = preds.reindex(columns=new_order)
            preds = preds.values.squeeze()
            preds = preds if prob_cols else target_enc.encode(preds).tolist()
            truths = target_enc.encode(truths).numpy()

            labels = self.compute_categorical_labels(preds, truths)

        else:
            raise(Exception(f"dtype {self.target_dtype} not supported for confidence normalizer"))

        return labels

    @staticmethod
    def compute_numerical_labels(preds: np.ndarray, truths: np.ndarray, bounds: list) -> np.ndarray:
        diffs = np.log(abs(preds - truths))
        diffs = diffs / np.max(diffs) if np.max(diffs) > 0 else diffs
        labels = np.clip(bounds[0] + diffs, bounds[0], bounds[1])
        return labels

    @staticmethod
    def compute_categorical_labels(preds: np.ndarray, truths: np.ndarray) -> np.ndarray:
        preds = np.clip(preds, 0.001, 0.999)  # avoid inf
        labels = entropy(truths, preds, axis=1)
        return labels



import time
import math
import torch
import numpy as np
import pandas as pd
import optuna
from typing import Dict, Union, Optional
from optuna import trial as trial_module
from sklearn import clone
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import check_cv, cross_val_predict
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier

from type_infer.dtype import dtype
from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments


class RandomForest(BaseMixer):
    model: Union[RandomForestClassifier, RandomForestRegressor]
    dtype_dict: dict
    target: str
    fit_on_dev: bool
    use_optuna: bool
    supports_proba: bool

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            fit_on_dev: bool,
            target_encoder: BaseEncoder,
            use_optuna: bool = False,
    ):
        """
        The `RandomForest` mixer supports both regression and classification tasks. 
        It inherits from sklearn.ensemble.RandomForestRegressor and sklearn.ensemble.RandomForestClassifier.
        (https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html)
        (https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html)

        :param stop_after: time budget in seconds.
        :param target: name of the target column that the mixer will learn to predict.
        :param dtype_dict: dictionary with dtypes of all columns in the data.
        :param fit_on_dev: whether to fit on the dev dataset.
        :param use_optuna: whether to activate the automated hyperparameter search (optuna-based). Note that setting this flag to `True` does not guarantee the search will run, rather, the speed criteria will be checked first (i.e., if a single iteration is too slow with respect to the time budget, the search will not take place). 
        """  # noqa
        super().__init__(stop_after)
        self.target = target
        self.dtype_dict = dtype_dict
        self.fit_on_dev = fit_on_dev
        self.use_optuna = use_optuna
        self.target_encoder = target_encoder

        self.model = None
        self.positive_domain = False
        self.num_trials = 5
        self.cv = 3
        self.map = {}

        self.cls_dtypes = [dtype.categorical, dtype.binary, dtype.cat_tsarray]
        self.float_dtypes = [dtype.float, dtype.quantity, dtype.num_tsarray]
        self.num_dtypes = [dtype.integer] + self.float_dtypes
        self.supports_proba = dtype_dict[target] in self.cls_dtypes
        self.is_classifier = self.supports_proba

        self.stable = True

    def _multi_logloss(self, y_true: np.ndarray, y_pred: np.ndarray, eps: float = 1e-15):
        # ('evaluate model effects' not  use this function)
        y_pred = np.clip(y_pred, eps, 1 - eps)
        score = np.mean([-math.log(y_pred[i][self.map[y]]) for i, y in enumerate(y_true)])

        return score

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the RandomForest model.

        :param train_data: encoded features for training dataset
        :param dev_data: encoded features for dev dataset
        """
        started = time.time()
        log.info('Started fitting RandomForest model')

        output_dtype = self.dtype_dict[self.target]

        if output_dtype not in self.cls_dtypes + self.num_dtypes:
            log.error(f'RandomForest mixer not supported for type: {output_dtype}')
            raise Exception(f'RandomForest mixer not supported for type: {output_dtype}')

        # concat the data if fit on dev
        if self.fit_on_dev:
            train_data = ConcatedEncodedDs([train_data, dev_data])

        # initialize the model
        init_params = {
            'n_estimators': 50,
            'max_depth': 5,
            'bootstrap': True,
            'n_jobs': -1,
            'random_state': 0
        }

        if self.is_classifier:
            X = train_data.get_encoded_data(include_target=False)
            Y = train_data.get_column_original_data(self.target)

            self.model = RandomForestClassifier(**init_params)

            self.model.fit(X, Y)  # sample_weight

            self.map = {cat: idx for idx, cat in enumerate(self.model.classes_)}  # for multi_logloss
        else:
            X = train_data.get_encoded_data(include_target=False)
            Y = train_data.get_encoded_column_data(self.target)

            self.model = RandomForestRegressor(**init_params)

            self.model.fit(X, Y)  # sample_weight

        # optimize params
        metric, predict_method = (self._multi_logloss, 'predict_proba') if self.is_classifier \
            else (mean_squared_error, 'predict')

        def objective(trial: trial_module.Trial):
            criterion = trial.suggest_categorical("criterion", "gini") if self.is_classifier else 'squared_error'

            params = {
                'n_estimators': trial.suggest_int('n_estimators', 2, 512),
                'criterion': criterion,
            }

            self.model.set_params(**params)

            y_pred = cross_val_predict(self.model, X, Y, cv=self.cv, method=predict_method)
            cv = check_cv(self.cv, Y, classifier=self.is_classifier)  #
            score = np.mean([metric(np.array(Y)[val_idx], y_pred[val_idx]) for _, val_idx in cv.split(X, Y)])

            return score

        elapsed = time.time() - started
        num_trials = max(min(int(self.stop_after / elapsed) - 2, self.num_trials), 0)
        if self.use_optuna:
            log.info(f'The number of trials (Optuna) is {num_trials}.')

        if self.use_optuna and num_trials > 0:
            init_score = metric(Y, getattr(self.model, predict_method)(X))

            study = optuna.create_study(direction='minimize')
            study.optimize(objective, n_trials=num_trials)

            opt_model = clone(self.model)
            opt_model.set_params(**study.best_params)
            opt_model.fit(X, Y)
            optuna_score = metric(Y, getattr(opt_model, predict_method)(X))
            log.info(f'init_score: {init_score}, optuna_score: {optuna_score}')

            if init_score <= optuna_score:
                self.model.set_params(**init_params)
            else:
                self.model = opt_model
                log.info(f'RandomForest parameters of the best trial: {study.best_params}')

        # evaluate model effects
        if self.fit_on_dev:
            log.info(f'RandomForest based correlation of (train data): {self.model.score(X, Y)}')
            X = dev_data.get_encoded_data(include_target=False)
            if self.is_classifier:
                Y = dev_data.get_column_original_data(self.target)
            else:
                Y = dev_data.get_encoded_column_data(self.target)
            log.info(f'RandomForest based correlation of (dev data): {self.model.score(X, Y)}')
        else:
            log.info(f'RandomForest based correlation of: {self.model.score(X, Y)}')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        """
        The RandomForest mixer does not support updates. If the model does not exist, a new one will be created and fitted. 

        :param train_data: encoded features for (new) training dataset
        :param dev_data: encoded features for (new) dev dataset
        """  # noqa
        if self.model is None:
            self.fit(train_data, dev_data)

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Call a trained RandomForest mixer to output predictions for the target column.

        :param ds: input data with values for all non-target columns.
        :param args: inference-time arguments (e.g. whether to output predicted labels or probabilities).

        :return: dataframe with predictions.
        """
        data = ds.get_encoded_data(include_target=False).numpy()

        if self.is_classifier:
            predictions = self.model.predict_proba(data)
            decoded_predictions = self.model.classes_.take(np.argmax(predictions, axis=1), axis=0)
        else:
            predictions = self.model.predict(data)
            if predictions.ndim == 1:
                decoded_predictions = predictions
            else:
                decoded_predictions = self.target_encoder.decode(torch.Tensor(predictions))

        if self.positive_domain:
            decoded_predictions = [max(0, p) for p in decoded_predictions]

        ydf = pd.DataFrame({'prediction': decoded_predictions})

        if args.predict_proba and self.supports_proba:
            for idx, label in enumerate(self.model.classes_):
                ydf[f'__mdb_proba_{label}'] = predictions[:, idx]

        return ydf



import torch
from torch import nn
from lightwood.mixer.helpers.default_net import DefaultNet
from lightwood.helpers.torch import LightwoodAutocast


class ArNet(DefaultNet):
    """
    DefaultNet variant that adds a secondary stream (simple linear layer) with constrained
    weights to learn autoregressive coefficients for numerical time series targets.
    """

    def __init__(self,
                 encoder_span: dict,  # contains index span for each encoder
                 target_name: str,
                 input_size: int = None,
                 output_size: int = None,
                 shape: list = None,
                 max_params: int = 3e7,
                 num_hidden: int = 1,
                 dropout: float = 0) -> None:

        self.ar_net = None
        super().__init__(input_size=input_size,
                         output_size=output_size,
                         shape=shape,
                         max_params=max_params,
                         num_hidden=num_hidden,
                         dropout=dropout
                         )
        self.target = target_name
        self.encoder_span = encoder_span
        self.ar_column = f'__mdb_ts_previous_{self.target}'
        self.ar_idxs = list(*[range(idx[0], idx[1]) for col, idx in encoder_span.items() if col == self.ar_column])
        dims = [(len(self.ar_idxs), output_size)]
        linears = [nn.Linear(in_features=inf, out_features=outf) for inf, outf in dims]
        self.ar_net = nn.Sequential(*linears)
        self.ar_net.to(self.device)

    def to(self, device=None, available_devices=None):
        if self.ar_net:
            self.ar_net.to(device)
        return super().to(device)

    def forward(self, input):
        with LightwoodAutocast():
            if len(input.shape) == 1:
                input = input.unsqueeze(0)

            residual_output = self.net(input)
            ar_output = self.ar_net(input[:, self.ar_idxs])
            if self.ar_net.training:
                self.ar_net._modules['0'].weight = nn.Parameter(torch.clamp(self.ar_net._modules['0'].weight,
                                                                            0.0,
                                                                            0.999))  # force unit root
        return ar_output + residual_output



from typing import Tuple, Dict, Optional

import pandas as pd
from lightwood.helpers.log import log


class BaseAnalysisBlock:
    """Class to be inherited by any analysis/explainer block."""
    def __init__(self,
                 deps: Optional[Tuple] = ()
                 ):

        self.dependencies = deps  # can be parallelized when there are no dependencies @TODO enforce

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        """
        This method should be called once during the analysis phase, or not called at all.
        It computes any information that the block may either output to the model analysis object,
        or use at inference time when `.explain()` is called (in this case, make sure all needed
        objects are added to the runtime analyzer so that `.explain()` can access them).

        :param info: Dictionary where any new information or objects are added. The next analysis block will use
        the output of the previous block as a starting point.
        :param kwargs: Dictionary with named variables from either the core analysis or the rest of the prediction
        pipeline.
        """
        log.info(f"{self.__class__.__name__}.analyze() has not been implemented, no modifications will be done to the model analysis.")  # noqa
        return info

    def explain(self,
                row_insights: pd.DataFrame,
                global_insights: Dict[str, object], **kwargs) -> Tuple[pd.DataFrame, Dict[str, object]]:
        """
        This method should be called once during the explaining phase at inference time, or not called at all.
        Additional explanations can be at an instance level (row-wise) or global.
        For the former, return a data frame with any new insights. For the latter, a dictionary is required.

        :param row_insights: dataframe with previously computed row-level explanations.
        :param global_insights: dict() with any explanations that concern all predicted instances or the model itself.

        :returns:
            - row_insights: modified input dataframe with any new row insights added here.
            - global_insights: dict() with any explanations that concern all predicted instances or the model itself.
        """
        log.info(f"{self.__class__.__name__}.explain() has not been implemented, no modifications will be done to the data insights.")  # noqa
        return row_insights, global_insights



# Original author: Henrik Linusson (github.com/donlnz)

import numpy as np


# -----------------------------------------------------------------------------
# Validity measures
# -----------------------------------------------------------------------------
def reg_n_correct(prediction, y, significance=None):
    """Calculates the number of correct predictions made by a conformal
    regression model.
    """
    if significance is not None:
        idx = int(significance * 100 - 1)
        prediction = prediction[:, :, idx]

    low = y >= prediction[:, 0]
    high = y <= prediction[:, 1]
    correct = low * high

    return y[correct].size


def reg_mean_errors(prediction, y, significance):
    """Calculates the average error rate of a conformal regression model.
    """
    return 1 - reg_n_correct(prediction, y, significance) / y.size


def class_n_correct(prediction, y, significance):
    """Calculates the number of correct predictions made by a conformal
    classification model.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    correct = np.zeros((y.size,), dtype=bool)
    for i, y_ in enumerate(y):
        correct[i] = prediction[i, int(y_)]
    return np.sum(correct)


def class_mean_errors(prediction, y, significance=None):
    """Calculates the average error rate of a conformal classification model.
    """
    return 1 - (class_n_correct(prediction, y, significance) / y.size)


def class_one_err(prediction, y, significance=None):
    """Calculates the error rate of conformal classifier predictions containing
     only a single output label.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    idx = np.arange(0, y.size, 1)
    idx = filter(lambda x: np.sum(prediction[x, :]) == 1, idx)
    errors = filter(lambda x: not prediction[x, int(y[x])], idx)

    if len(idx) > 0:
        return np.size(errors) / np.size(idx)
    else:
        return 0


def class_mean_errors_one_class(prediction, y, significance, c=0):
    """Calculates the average error rate of a conformal classification model,
      considering only test examples belonging to class ``c``. Use
      ``functools.partial`` in order to test other classes.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    idx = np.arange(0, y.size, 1)[y == c]
    errs = np.sum(1 for _ in filter(lambda x: not prediction[x, c], idx))

    if idx.size > 0:
        return errs / idx.size
    else:
        return 0


def class_one_err_one_class(prediction, y, significance, c=0):
    """Calculates the error rate of conformal classifier predictions containing
     only a single output label. Considers only test examples belonging to
     class ``c``. Use ``functools.partial`` in order to test other classes.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    idx = np.arange(0, y.size, 1)
    idx = filter(lambda x: prediction[x, c], idx)
    idx = filter(lambda x: np.sum(prediction[x, :]) == 1, idx)
    errors = filter(lambda x: int(y[x]) != c, idx)

    if len(idx) > 0:
        return np.size(errors) / np.size(idx)
    else:
        return 0


# -----------------------------------------------------------------------------
# Efficiency measures
# -----------------------------------------------------------------------------
def _reg_interval_size(prediction, y, significance):
    idx = int(significance * 100 - 1)
    prediction = prediction[:, :, idx]

    return prediction[:, 1] - prediction[:, 0]


def reg_min_size(prediction, y, significance):
    return np.min(_reg_interval_size(prediction, y, significance))


def reg_q1_size(prediction, y, significance):
    return np.percentile(_reg_interval_size(prediction, y, significance), 25)


def reg_median_size(prediction, y, significance):
    return np.median(_reg_interval_size(prediction, y, significance))


def reg_q3_size(prediction, y, significance):
    return np.percentile(_reg_interval_size(prediction, y, significance), 75)


def reg_max_size(prediction, y, significance):
    return np.max(_reg_interval_size(prediction, y, significance))


def reg_mean_size(prediction, y, significance):
    """Calculates the average prediction interval size of a conformal
    regression model.
    """
    return np.mean(_reg_interval_size(prediction, y, significance))


def class_avg_c(prediction, y, significance):
    """Calculates the average number of classes per prediction of a conformal
    classification model.
    """
    prediction = prediction > significance
    return np.sum(prediction) / prediction.shape[0]


def class_mean_p_val(prediction, y, significance):
    """Calculates the mean of the p-values output by a conformal classification
    model.
    """
    return np.mean(prediction)


def class_one_c(prediction, y, significance):
    """Calculates the rate of singleton predictions (prediction sets containing
    only a single class label) of a conformal classification model.
    """
    prediction = prediction > significance
    n_singletons = np.sum(1 for _ in filter(lambda x: np.sum(x) == 1, prediction))
    return n_singletons / y.size


def class_empty(prediction, y, significance):
    """Calculates the rate of singleton predictions (prediction sets containing
    only a single class label) of a conformal classification model.
    """
    prediction = prediction > significance
    n_empty = np.sum(1 for _ in filter(lambda x: np.sum(x) == 0, prediction))
    return n_empty / y.size


def n_test(prediction, y, significance):
    """Provides the number of test patters used in the evaluation.
    """
    return y.size



from typing import List, Union
from collections import defaultdict
import pandas as pd
import numpy as np
import torch

from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.constants import _UNCOMMON_WORD


class SimpleLabelEncoder(BaseEncoder):
    """
    Simple encoder that assigns a unique integer to every observed label.
    
    Allocates an `unknown` label by default to index 0.
    
    Labels must be exact matches between inference and training (e.g. no .lower() on strings is performed here).
    """  # noqa

    def __init__(self, is_target=False, normalize=True) -> None:
        super().__init__(is_target)
        self.label_map = defaultdict(int)  # UNK category maps to 0
        self.inv_label_map = {}  # invalid encoded values are mapped to None in `decode`
        self.output_size = 1
        self.n_labels = None
        self.normalize = normalize

    def prepare(self, priming_data: Union[list, pd.Series]) -> None:
        if not isinstance(priming_data, pd.Series):
            priming_data = pd.Series(priming_data)

        for i, v in enumerate(priming_data.unique()):
            if v is not None:
                self.label_map[str(v)] = int(i + 1)  # leave 0 for UNK
        self.n_labels = len(self.label_map)
        for k, v in self.label_map.items():
            self.inv_label_map[v] = k
        self.is_prepared = True

    def encode(self, data: Union[tuple, np.ndarray, pd.Series], normalize=True) -> torch.Tensor:
        """
        :param normalize: can be used to temporarily return unnormalized values
        """
        if not isinstance(data, pd.Series):
            data = pd.Series(data)  # specific to the Gym class - remove once deprecated!
        if isinstance(data, np.ndarray):
            data = pd.Series(data)

        data = data.astype(str)
        encoded = torch.Tensor(data.map(self.label_map))

        if normalize and self.normalize:
            encoded /= self.n_labels
        if len(encoded.shape) < 2:
            encoded = encoded.unsqueeze(-1)

        return encoded

    def decode(self, encoded_values: torch.Tensor, normalize=True) -> List[object]:
        """
        :param normalize: can be used to temporarily return unnormalized values
        """
        if normalize and self.normalize:
            encoded_values *= self.n_labels
        values = encoded_values.long().squeeze().tolist()  # long() as inv_label_map expects an int key
        values = [self.inv_label_map.get(v, _UNCOMMON_WORD) for v in values]
        return values
